# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Module for setting paths used throughout GeoIPS.

This module sets various directory paths required for GeoIPS.
Setting these variables here keeps geoips code and outputs consolidated.
Eventually, we aim to supplant this method of setting variables with a config file.
`GEOIPS_OUTDIRS` serves as the base reference for all other variables and defaults
to the current directory.

The module defaults to the values set in environment variables for all other variables.
It also provides a function for creating directories.

Functions
---------
- get_env_var: Retrieve an environment variable or a provided default value.
- initialize_paths: Returns a dictionary with GeoIPS directories and variables.
- make_dirs: Create directories if they don't already exist.

Attributes
----------
- PATHS: Dictionary with initialized paths for various GeoIPS directories and URLs.

Environment Variables:
- GEOIPS_OUTDIRS (required): Base output directory for GeoIPS.
"""

import logging
import os
import socket

LOG = logging.getLogger(__name__)


def get_env_var(var_name, default, rstrip_path=True):
    """Retrieve environment variable or provided default, optionally rstrip a '/'.

    Parameters
    ----------
    var_name : str
        The name of the environment variable.
    default : str
        The default value to return if the environment variable is not set.
    rstrip_path : bool, optional
        If True, strip trailing slashes from the returned path (default is True).

    Returns
    -------
    str
        The value of the environment variable or the default value if not set.
    """
    env_value = os.getenv(var_name)
    if env_value:
        if rstrip_path:
            return env_value.rstrip("/")
        else:
            return env_value
    else:
        return default


def initialize_paths():
    """
    Initialize and return a dictionary of paths used throughout GeoIPS.

    Returns
    -------
    dict
        A dictionary containing various paths used by GeoIPS.

    Raises
    ------
    KeyError
        If the required environment variable GEOIPS_OUTDIRS is not set.
    """
    paths = {}

    # Base and Documentation Paths
    paths["BASE_PATH"] = os.path.abspath(
        os.path.join(os.path.dirname(__file__), os.pardir)
    )

    # Output Directories
    try:
        paths["GEOIPS_OUTDIRS"] = os.environ["GEOIPS_OUTDIRS"].rstrip("/")
    except KeyError:
        LOG.warning(
            "GEOIPS_OUTDIRS is not set in your environment. "
            "Defaulting to $HOME/GEOIPS_OUTDIRS as output dir. "
            "Please set GEOIPS_OUTDIRS if you want to write to a "
            "different directory by default."
        )
        try:
            home_dir = os.environ["HOME"]
        except KeyError:
            LOG.error(
                "Could not resolve enviorment variable " "'$HOME', using '/' as default"
            )
            home_dir = "/"
        paths["GEOIPS_OUTDIRS"] = os.path.join(home_dir, "GEOIPS_OUTDIRS")
    paths["GEOIPS_PACKAGES_DIR"] = os.path.abspath(
        os.path.join(paths["BASE_PATH"], os.pardir, os.pardir)
    )
    paths["GEOIPS_BASEDIR"] = get_env_var(
        "GEOIPS_BASEDIR",
        os.path.abspath(os.path.join(paths["GEOIPS_PACKAGES_DIR"], os.pardir)),
    )
    paths["GEOIPS_REBUILD_REGISTRIES"] = get_env_var(
        "GEOIPS_REBUILD_REGISTRIES",
        True,
    )
    # Convert the string to a bool
    if paths["GEOIPS_REBUILD_REGISTRIES"] == "0":
        paths["GEOIPS_REBUILD_REGISTRIES"] = False

    # Identify defaults for global GeoIPS variables.  The actual values for
    # these variables will be set using get_env_var below.
    geoips_global_variable_defaults = {
        # GeoIPS Documentation URL
        "GEOIPS_DOCS_URL": r"https://nrlmmd-geoips.github.io/geoips/",
        # Version
        "GEOIPS_VERS": "0.0.0",
        # Operational User
        "GEOIPS_OPERATIONAL_USER": False,
        # Copyright Information
        "GEOIPS_COPYRIGHT": "NRL-Monterey",
        "GEOIPS_COPYRIGHT_ABBREVIATED": "NRLMRY",
        # Configuration and Queue
        "GEOIPS_RCFILE": "",
        "DEFAULT_QUEUE": None,
        # Computer Identifier
        "BOXNAME": socket.gethostname(),
        # Threshold for image-based output checks.  This will be cast to float below.
        "OUTPUT_CHECKER_THRESHOLD_IMAGE": 0.05,
        # Minimal default
        # "GEOIPS_LOG_FMT_STRING": "%(asctime)s: %(message)s"
        # "GEOIPS_LOG_DATEFMT_STRING": "%d_%H%M%S"
        # More informative default
        "GEOIPS_LOGGING_FMT_STRING": "%(asctime)s %(module)12s.py:%(lineno)-4d "
        "%(levelname)7s: %(message)s",
        "GEOIPS_LOGGING_DATEFMT_STRING": "%d_%H%M%S",
        "GEOIPS_LOGGING_LEVEL": "interactive",
    }

    # Identify the defaults for relative path-based environment variables,
    # these paths default to locations under base paths set above.
    # The actual variables will be set using get_env_var below.
    default_derivative_directory_path_defaults = {
        paths["GEOIPS_BASEDIR"]: {
            "GEOIPS_TESTDATA_DIR": "test_data",
            "GEOIPS_DEPENDENCIES_DIR": "geoips_dependencies",
        },
        paths["GEOIPS_OUTDIRS"]: {
            # Data Output Paths
            "PRESECTORED_DATA_PATH": "preprocessed/sectored",
            "PREREAD_DATA_PATH": "preprocessed/unsectored",
            "PREREGISTERED_DATA_PATH": "preprocessed/registered",
            "PRECALCULATED_DATA_PATH": "preprocessed/algorithms",
            "CLEAN_IMAGERY_PATH": "preprocessed/clean_imagery",
            "ANNOTATED_IMAGERY_PATH": "preprocessed/annotated_imagery",
            "GEOTIFF_IMAGERY_PATH": "preprocessed/geotiff_imagery",
            "FINAL_DATA_PATH": "preprocessed/final",
            "PREGENERATED_GEOLOCATION_PATH": "preprocessed/geolocation",
            # Scratch Directories
            "SCRATCH": "scratch",
            "LOCALSCRATCH": "scratch",
            "SHAREDSCRATCH": "scratch",
            # Log and Data Directories
            "LOGDIR": "logs",
            "GEOIPSDATA": "geoipsdata",
            # Ancillary Data Directories
            "GEOIPS_ANCILDAT_AUTOGEN": "ancildat_autogen",
            "GEOIPS_ANCILDAT": "ancildat",
            # WWW Paths
            "TCWWW": "preprocessed/tcwww",
            "PUBLICWWW": "preprocessed/publicwww",
            "PRIVATEWWW": "preprocessed/privatewww",
            # Tropical Cyclone Paths
            "TC_DECKS_DB": "longterm_files/tc/tc_decks.db",
            "TC_DECKS_DIR": "longterm_files/tc/decks",
        },
        paths["BASE_PATH"]: {
            "TC_TEMPLATE": "plugins/yaml/sectors/dynamic/tc_web_template.yaml",
        },
        paths["GEOIPS_REBUILD_REGISTRIES"]: {
            "GEOIPS_REBUILD_REGISTRIES_TRUE": True,
            "GEOIPS_REBUILD_REGISTRIES_FALSE": False,
        },
    }

    # looping through all the directory-based paths and global variables set above
    # using "get_env_var" function to set the variables to the environment variable
    # specified option (when defined via the first argument)
    # else defaulting to the passed-in default (second argument)
    for key, value in geoips_global_variable_defaults.items():
        paths[key] = get_env_var(key, value)

    for (
        top_directory,
        sub_directories,
    ) in default_derivative_directory_path_defaults.items():
        for key, sub_path in sub_directories.items():
            if "GEOIPS_REBUILD_REGISTRIES" in key:
                paths[key] = get_env_var(key, sub_path, rstrip_path=False)
            else:
                paths[key] = get_env_var(key, os.path.join(top_directory, sub_path))

    # Handling special cases now: home for linux/windows
    if not os.getenv("HOME"):
        # need home drive default for windows
        paths["HOME"] = os.getenv("HOMEDRIVE") + os.getenv("HOMEPATH")
    else:
        paths["HOME"] = os.getenv("HOME").rstrip("/")

    # Setting links for WWW Paths
    www_paths = ["TCWWW", "PUBLICWWW", "PRIVATEWWW"]
    for path in www_paths:
        paths[f"{path}_URL"] = get_env_var(f"{path}_URL", paths[path])

    # This needs to be a float
    paths["OUTPUT_CHECKER_THRESHOLD_IMAGE"] = float(
        paths["OUTPUT_CHECKER_THRESHOLD_IMAGE"]
    )

    return paths


def make_dirs(path):
    """Make directories, catching exceptions if directory already exists.

    Parameters
    ----------
    path : str
        Path to directory to create

    Returns
    -------
    str
        Path if successfully created
    """
    LOG.info("Creating directory %s if it doesn't already exist.", path)
    os.makedirs(path, mode=0o755, exist_ok=True)
    return path


# Initialize the PATHS dictionary
PATHS = initialize_paths()



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""TestPluginRegistry Class used for Unit Testing the Plugin Registries."""

from glob import glob
from importlib import import_module, metadata
import logging
from os.path import basename, splitext
from pprint import pformat
import pytest
import json
import yaml

from geoips.errors import PluginError, PluginRegistryError
from geoips.filenames.base_paths import PATHS
from geoips.interfaces import algorithms, products, sectors, workflows
from geoips.plugin_registry import PluginRegistry

LOG = logging.getLogger(__name__)


class FakeInterface:
    """Dummy fake interface used to cause appropriate errors from the PluginRegistry."""

    interface_type = "fake"
    name = "fake_interface"


class PluginRegistryValidator(PluginRegistry):
    """Subclass of PluginRegistry which adds functionality for unit testing."""

    def __init__(self, fpaths=None):
        """Initialize TestPluginRegistry Class."""
        super().__init__("geoips.plugin_packages", _test_registry_files=fpaths)

    def validate_plugin_types_exist(self, reg_dict, reg_path):
        """Test that all top level plugin types exist in each registry file."""
        expected_plugin_types = ["yaml_based", "module_based", "text_based"]
        for p_type in expected_plugin_types:
            if p_type not in reg_dict:
                error_str = f"Expected plugin type '{p_type}' to be in the registry but"
                error_str += f" wasn't. This was in file '{reg_path}'."
                raise PluginRegistryError(error_str)

    def validate_all_registries(self):
        """Validate all registries in the current installation.

        This should be run during testing, but not at runtime.
        Ensure we do not fail catastrophically for a single bad plugin
        at runtime, so test up front to test validity.
        """
        for reg_path in self.registry_files:
            pkg_plugins = json.load(open(reg_path, "r"))
            self.validate_registry(pkg_plugins, reg_path)

    def validate_registry(self, current_registry, fpath):
        """Test all plugins found in registered plugins for their validity."""
        try:
            self.validate_plugin_types_exist(current_registry, fpath)
        except PluginRegistryError as e:
            # xfail if this is a test, otherwise just raise PluginRegistryError
            if self._is_test:
                pytest.xfail(str(e))
            else:
                raise PluginRegistryError(e)
        try:
            self.validate_registry_interfaces(current_registry)
        except PluginRegistryError as e:
            # xfail if this is a test, otherwise just raise PluginRegistryError
            if self._is_test:
                pytest.xfail(str(e))
            else:
                raise PluginRegistryError(e)
        for plugin_type in current_registry:
            for interface in current_registry[plugin_type]:
                for plugin in current_registry[plugin_type][interface]:
                    try:
                        if interface == "products":
                            for subplg in current_registry[plugin_type][interface][
                                plugin
                            ]:
                                self.validate_plugin_attrs(
                                    plugin_type,
                                    interface,
                                    (plugin, subplg),
                                    current_registry[plugin_type][interface][plugin][
                                        subplg
                                    ],
                                )
                        elif plugin_type != "text_based":
                            self.validate_plugin_attrs(
                                plugin_type,
                                interface,
                                plugin,
                                current_registry[plugin_type][interface][plugin],
                            )
                    except PluginRegistryError as e:
                        # xfail if this is a test,
                        # otherwise just raise PluginRegistryError
                        if self._is_test:
                            pytest.xfail(str(e))
                        else:
                            raise PluginRegistryError(e)

    def validate_plugin_attrs(self, plugin_type, interface, name, plugin):
        """Test non-product plugin for all required attributes."""
        missing = []
        if plugin_type == "yaml_based" and interface != "products":
            attrs = [
                "docstring",
                "family",
                "interface",
                "package",
                "plugin_type",
                "relpath",
            ]
        elif plugin_type == "yaml_based":
            attrs = [
                "docstring",
                "family",
                "interface",
                "package",
                "plugin_type",
                "product_defaults",
                "source_names",
                "relpath",
            ]
        else:
            attrs = [
                "docstring",
                "family",
                "interface",
                "package",
                "plugin_type",
                "signature",
                "relpath",
            ]
        for attr in attrs:
            try:
                plugin[attr]
            except KeyError:
                missing.append(attr)
        if missing:
            raise PluginRegistryError(
                f"Plugin '{name}' is missing the following required "
                f"top-level properties: '{missing}'"
            )

    def validate_registry_interfaces(self, current_registry):
        """Test Plugin Registry interfaces validity."""
        yaml_interfaces = []
        module_interfaces = []

        # NOTE: all <pkg>/interfaces/__init__.py files MUST
        # contain a "module_based_interfaces" list and a
        # "yaml_based_interfaces" list - this is how we
        # identify the valid interface names.
        # We must avoid actually importing the interfaces within
        # the geoips repo - or we will end up with a circular import
        # due to BaseInterface.
        for pkg in metadata.entry_points(group="geoips.plugin_packages"):
            try:
                mod = import_module(f"{pkg.value}.interfaces")
            except ModuleNotFoundError as resp:
                if f"No module named '{pkg.value}.interfaces'" in str(resp):
                    continue
                else:
                    raise ModuleNotFoundError(resp)
            module_interfaces += mod.module_based_interfaces
            yaml_interfaces += mod.yaml_based_interfaces

        bad_interfaces = []
        for plugin_type in ["module_based", "yaml_based"]:
            for interface in current_registry[plugin_type]:
                if (
                    interface not in module_interfaces
                    and interface not in yaml_interfaces
                ):
                    error_str = f"\nPlugin type '{plugin_type}' does not allow "
                    error_str += f"interface '{interface}'.\n"
                    error_str += "\nValid interfaces: "
                    if plugin_type == "module_based":
                        interface_list = module_interfaces
                    else:
                        interface_list = yaml_interfaces
                    error_str += f"\n{interface_list}\n"
                    error_str += "\nPlease update the following plugins "
                    error_str += "to use a valid interface:\n"
                    error_str += pformat(current_registry[plugin_type][interface])
                    bad_interfaces.append(error_str)
        if bad_interfaces:
            error_str = "The following interfaces were not valid:\n"
            for error in bad_interfaces:
                error_str += error
            raise PluginRegistryError(error_str)


class TestPluginRegistry:
    """
    Pytest-based Unit Test for the PluginRegistry Class.

    Note: Since we are not able to initialize this class due to restrictions placed by
    Pytest, if you want to change the test files used, simply replace the location below
    with the location of your new test files.
    """

    default_fpaths = glob(
        str(__file__).replace("test_plugin_registries.py", "files/**/*.yaml"),
        recursive=True,
    )

    # pr_validator uses test registry files for most tests
    pr_validator = PluginRegistryValidator(fpaths=default_fpaths)
    # real reg validator is responsible for testing deleting and building of registries
    # based on whether or not we want that to occur:
    # I.e. whether or not PATHS["GEOIPS_REBUILD_REGISTRIES"] is set to True or False
    real_reg_validator = PluginRegistryValidator(fpaths=None)

    # Couldn't implement this class via inheritance because PyTest Raised this error:
    # PytestCollectionWarning: cannot collect test class 'TestPluginRegistry' because
    # it has a __init__ constructor (from: test_plugin_registries.py)

    # This isn't a problem though, as we can just set it as a class attribute and move
    # forward as usual. Pytest is just picky about having __init__ in Test Classes.

    # def __init__(self, fpaths=default_fpaths):
    #     super().__init__(fpaths)

    def generate_id(fpath):
        """Generate pytest id for current test."""
        test_id = "bad"
        if "/good/" in fpath:
            test_id = "good"
        return f"{test_id}-{splitext(basename(fpath))[0]}"

    def test_registered_plugin_property(self):
        """Ensure registered_plugins is valid in its nature."""
        if hasattr(self.pr_validator, "_registered_plugins"):
            del self.pr_validator._registered_plugins
        print(self.pr_validator.registered_plugins)
        assert isinstance(self.pr_validator.registered_plugins, dict)
        assert "yaml_based" in self.pr_validator.registered_plugins
        assert "module_based" in self.pr_validator.registered_plugins
        assert "text_based" in self.pr_validator.registered_plugins
        assert self.pr_validator.registered_plugins

    def test_interface_mapping_property(self):
        """Ensure interface_mapping is valid in its nature."""
        if hasattr(self.pr_validator, "_interface_mapping"):
            del self.pr_validator._interface_mapping
        if hasattr(self.pr_validator, "_registered_plugins"):
            del self.pr_validator._registered_plugins
        print(self.pr_validator.interface_mapping)
        assert isinstance(self.pr_validator.interface_mapping, dict)
        assert "yaml_based" in self.pr_validator.interface_mapping
        assert "module_based" in self.pr_validator.interface_mapping
        assert "text_based" in self.pr_validator.interface_mapping
        assert isinstance(self.pr_validator.interface_mapping["yaml_based"], list)
        assert isinstance(self.pr_validator.interface_mapping["module_based"], list)
        assert isinstance(self.pr_validator.interface_mapping["text_based"], list)

    def test_registered_module_plugins_property(self):
        """Ensure that registered_module_plugins exist and have contents."""
        if hasattr(self.pr_validator, "_registered_plugins"):
            del self.pr_validator._registered_plugins
        print(self.pr_validator.registered_module_based_plugins)
        assert isinstance(self.pr_validator.registered_module_based_plugins, dict)
        assert len(self.pr_validator.registered_module_based_plugins)

    def test_registered_yaml_plugins_property(self):
        """Ensure that registered_yaml_plugins exist and have contents."""
        if hasattr(self.pr_validator, "_registered_plugins"):
            del self.pr_validator._registered_plugins
        print(self.pr_validator.registered_yaml_based_plugins)
        assert isinstance(self.pr_validator.registered_yaml_based_plugins, dict)
        assert len(self.pr_validator.registered_yaml_based_plugins)

    def test_automatic_registry_creation(self):
        """Assert that the registries are automatically rebuilt.

        This occurs if expected registry files are missing and
        PATHS["GEOIPS_REBUILD_REGISTRIES"] is set to True.
        """
        PATHS["GEOIPS_REBUILD_REGISTRIES"] = True
        self.real_reg_validator.delete_registries()
        # Delete again using specific packages (won't do anything) for code coverage
        self.real_reg_validator.delete_registries(packages=["geoips"])
        self.real_reg_validator._set_class_properties(force_reset=True)
        assert self.real_reg_validator.registered_plugins

    def test_disabled_registry_creation(self):
        """Assert that the registries are not automatically rebuilt.

        This occurs if expected registry files are missing and
        PATHS["GEOIPS_REBUILD_REGISTRIES"] is set to False. A FileNotFoundError should
        be raised.
        """
        PATHS["GEOIPS_REBUILD_REGISTRIES"] = False
        self.real_reg_validator.delete_registries()
        with pytest.raises(FileNotFoundError):
            self.real_reg_validator._set_class_properties(force_reset=True)

    def test_get_plugin_metadata(self):
        """Retrieve plugin metadata from a plugin that we know exists and is correct.

        For this test, we'll be using the algorithms.single_channel plugin and the
        products.abi.Infrared plugin.
        """
        PATHS["GEOIPS_REBUILD_REGISTRIES"] = True
        self.real_reg_validator._set_class_properties(force_reset=True)
        assert algorithms.get_plugin_metadata("single_channel")
        assert products.get_plugin_metadata("abi", "Infrared")

    def test_get_plugin_metadata_failing_cases(self):
        """Attempt to retrieve plugin metadata from a cases that we know should fail."""
        PATHS["GEOIPS_REBUILD_REGISTRIES"] = True
        self.real_reg_validator._set_class_properties(force_reset=True)
        # Caused when a plugin doesn't exist under an interface's registry
        with pytest.raises(PluginRegistryError):
            algorithms.get_plugin_metadata("fake_plugin")
        # Caused due to invalid argument type
        with pytest.raises(TypeError):
            algorithms.get_plugin_metadata(1078)

        # Caused due to the registry being unable to locate this inteface of a certain
        # type
        with pytest.raises(KeyError):
            self.real_reg_validator.get_plugin_metadata(FakeInterface, "fake_plugin")

    def test_get_yaml_plugin(self):
        """Retrieve valid (existing and formatted correctly) GeoIPS YAML plugin(s).

        Will retrieve ('abi', 'Infrared') product, 'goes_east' sector, and
        'abi_infrared' workflow plugins, as those all hit different portions of the
        codebase.
        """
        prd = self.real_reg_validator.get_yaml_plugin(products, ("abi", "Infrared"))
        sect = self.real_reg_validator.get_yaml_plugin(sectors, "goes_east")
        wrkflw = self.real_reg_validator.get_yaml_plugin(workflows, "abi_infrared")

        assert prd["name"] == "Infrared"
        assert prd["interface"] == "products"
        assert "abi" in prd["source_names"]

        assert sect["name"] == "goes_east"
        assert sect["interface"] == "sectors"

        assert wrkflw["name"] == "abi_infrared"
        assert wrkflw["interface"] == "workflows"

    def test_get_yaml_plugin_failing_cases(self):
        """Attempt to get all plugins from an interface using cases that should fail."""
        # Reconstruct the registry in memory so we start at a clean slate
        self.real_reg_validator._set_class_properties(force_reset=True)
        # Set interfaces' rebuild_registries attr to false for the time being
        sectors.rebuild_registries = False
        products.rebuild_registries = False
        workflows.rebuild_registries = False

        yam_reg = self.real_reg_validator.registered_plugins.pop("yaml_based")
        # Caused due to 'yaml_based' plugins being absent from the registry
        with pytest.raises(PluginError):
            self.real_reg_validator.get_yaml_plugin(sectors, "goes_east")
        # Reset the yaml_based portion of the registry back to its original value
        self.real_reg_validator.registered_plugins["yaml_based"] = yam_reg

        # Caused due to invalid argument type (rebuild_registries) should be a boolean
        with pytest.raises(TypeError):
            self.real_reg_validator.get_yaml_plugin(
                sectors, "goes_east", rebuild_registries=Exception
            )

        # Caused due to a missing plugin
        with pytest.raises(PluginError):
            self.real_reg_validator.get_yaml_plugin(
                products, ("abi", "fake_plugin"), rebuild_registries=False
            )

        abi_infrared_relpath = self.real_reg_validator.registered_yaml_based_plugins[
            "products"
        ]["abi"]["Infrared"].pop("relpath")
        self.real_reg_validator.registered_yaml_based_plugins["products"]["abi"][
            "Infrared"
        ]["relpath"] = "/some/fake/path.yaml"
        # Caused due to non existent file path
        with pytest.raises(PluginRegistryError):
            self.real_reg_validator.get_yaml_plugin(products, ("abi", "Infrared"))
        # Reset that relative path back to its original value
        self.real_reg_validator.registered_yaml_based_plugins["products"]["abi"][
            "Infrared"
        ]["relpath"] = abi_infrared_relpath

        fake_plugin_entry = {
            "docstring": "The 89VNearest product_defaults for gmi product.",
            "family": None,
            "interface": "products",
            "package": "geoips",
            "plugin_type": "yaml_based",
            "product_defaults": "89VNearest",
            "source_names": ["gmi"],
            "relpath": "plugins/yaml/products/gmi.yaml",
        }
        self.real_reg_validator.registered_plugins["yaml_based"]["products"]["gmi"][
            "fake_plugin"
        ] = fake_plugin_entry
        # Caused due to products plugin being found in the registry but doesn't exist
        # in the actual file specified
        with pytest.raises(PluginError):
            self.real_reg_validator.get_yaml_plugin(products, ("gmi", "fake_plugin"))
        # Remove the fake entry from the registry
        self.real_reg_validator.registered_plugins["yaml_based"]["products"]["gmi"].pop(
            "fake_plugin"
        )

        # Caused due to a missing plugin (that's not a products plugin)
        with pytest.raises(PluginError):
            self.real_reg_validator.get_yaml_plugin(sectors, "fake_plugin")

        goes_east_relpath = self.real_reg_validator.registered_yaml_based_plugins[
            "sectors"
        ]["goes_east"].pop("relpath")
        self.real_reg_validator.registered_yaml_based_plugins["sectors"]["goes_east"][
            "relpath"
        ] = "/some/fake/path.yaml"
        # Caused due to non existent file path
        with pytest.raises(PluginRegistryError):
            self.real_reg_validator.get_yaml_plugin(sectors, "goes_east")
        # Reset that relative path back to its original value
        self.real_reg_validator.registered_yaml_based_plugins["sectors"]["goes_east"][
            "relpath"
        ] = goes_east_relpath

        fake_plugin_entry = {
            "docstring": "ABI DWM High Workflow.",
            "family": "order_based",
            "interface": "workflows",
            "package": "geoips",
            "plugin_type": "yaml_based",
            "relpath": "plugins/yaml/workflows/abi.yaml",
        }
        self.real_reg_validator.registered_plugins["yaml_based"]["workflows"][
            "fake_plugin"
        ] = fake_plugin_entry
        # Caused due to workflow plugin being present in registry but not in
        # corresponding plugin
        with pytest.raises(PluginError):
            self.real_reg_validator.get_yaml_plugin(workflows, "fake_plugin")
        # Remove the fake entry from the registry
        self.real_reg_validator.registered_plugins["yaml_based"]["workflows"].pop(
            "fake_plugin"
        )
        # This test misses line #426 as that is a pydantic check. Can't test this
        # until we switch over to that.
        # Reset interfaces' rebuild_registries values to True
        products.rebuild_registries = True
        sectors.rebuild_registries = True
        workflows.rebuild_registries = True

    def test_get_yaml_plugins(self):
        """Retrieve valid (existing and formatted correctly) GeoIPS YAML plugins.

        This tests PluginRegistry.get_yaml_plugins, in this case, using the sectors
        interface.
        """
        self.real_reg_validator._set_class_properties(force_reset=True)
        assert len(self.real_reg_validator.get_yaml_plugins(sectors))

    def test_get_yaml_plugins_failing_cases(self):
        """Attempt to retrieve all yaml plugins from a fake interface."""
        plugins = self.real_reg_validator.get_yaml_plugins(FakeInterface)
        assert len(plugins) == 0

    def test_get_module_plugin(self):
        """Retrieve a valid (existing and formatted correctly) module plugin.

        In this test case, we are using algorithms.single_channel.
        """
        self.real_reg_validator._set_class_properties(force_reset=True)
        alg = self.real_reg_validator.get_module_plugin(algorithms, "single_channel")
        assert alg.name == "single_channel"
        assert alg.interface == "algorithms"

    def test_get_module_plugin_failing_cases(self):
        """Attempt to retrieve a module plugin using cases that we know will fail."""
        # Reconstruct the registry in memory so we start at a clean slate
        self.real_reg_validator._set_class_properties(force_reset=True)
        # Set algorithms' rebuild_registry attr to false for the time being
        algorithms.rebuild_registries = False
        mod_reg = self.real_reg_validator.registered_plugins.pop("module_based")
        # Caused due to 'module_based' not being at the top level of the registry
        with pytest.raises(PluginError):
            self.real_reg_validator.get_module_plugin(algorithms, "single_channel")
        # Reset the registry to its original state
        self.real_reg_validator.registered_plugins["module_based"] = mod_reg

        # Caused due to invalid argument type (rebuild_registries) should be a boolean
        with pytest.raises(TypeError):
            self.real_reg_validator.get_module_plugin(
                algorithms, "single_channel", rebuild_registries=Exception
            )

        # Caused due to a missing plugin
        with pytest.raises(PluginError):
            self.real_reg_validator.get_module_plugin(
                algorithms, "fake_plugin", rebuild_registries=False
            )

        fake_plugin_entry = {
            "docstring": (
                "Data manipulation steps for standard 'single_channel' algorithm.\n"
                "Generalized algorithm to apply data manipulation steps in a standard "
                "order to apply corrections to a single channel output product."
            ),
            "family": "list_numpy_to_numpy",
            "interface": "algorithms",
            "package": "geoips",
            "plugin_type": "module_based",
            "signature": (
                "(arrays, output_data_range=None, input_units=None, output_units=None, "
                "min_outbounds='crop', max_outbounds='crop', norm=False, inverse=False,"
                " sun_zen_correction=False, mask_night=False, max_day_zen=None, "
                "mask_day=False, min_night_zen=None, gamma_list=None, "
                "scale_factor=None, satellite_zenith_angle_cutoff=None)"
            ),
            "relpath": "/some/fake/path.py",
        }
        self.real_reg_validator.registered_plugins["module_based"]["algorithms"][
            "fake_plugin"
        ] = fake_plugin_entry
        # Caused due to algorithm plugin being present in registry it's relative path
        # does not exist
        with pytest.raises(PluginRegistryError):
            self.real_reg_validator.get_module_plugin(algorithms, "fake_plugin")
        # Remove the fake entry from the registry
        self.real_reg_validator.registered_plugins["module_based"]["algorithms"].pop(
            "fake_plugin"
        )
        # reset algorithms' rebuild_registries attr to true
        algorithms.rebuild_registries = True

    def test_get_module_plugins(self):
        """Retrieve valid (existing and formatted correctly) module plugins.

        In this case, we are retrieving all GeoIPS module algorithm plugins.
        """
        self.real_reg_validator._set_class_properties(force_reset=True)
        assert len(self.real_reg_validator.get_module_plugins(algorithms))

    def test_get_module_plugins_failing_cases(self):
        """Attempt to retrieve all module plugins from a fake interface."""
        plugins = self.real_reg_validator.get_module_plugins(FakeInterface)
        assert len(plugins) == 0

    def test_retry_get_plugin(self):
        """Test that PluginRegistry.retry_get_plugin works as expected."""
        with pytest.raises(PluginError):
            self.real_reg_validator.get_module_plugin(
                algorithms, "fake_plugin", rebuild_registries=True
            )

    def test_create_registries_invalid_input(self):
        """Assert ValueError is raised when invalid input is sent to create_registries.

        Where 'create_registries' comes from PluginRegistry.
        """
        # Caused due to invalid input. save_type must be one of ['json', 'yaml']
        with pytest.raises(ValueError):
            self.real_reg_validator.create_registries(save_type="log")

    def test_create_registries_with_specific_packages(self, caplog):
        """Test that create_registries works with a subset of packages provided.

        In this case, packages = ['geoips'].
        """
        # 10 is the level of LOG.debug
        with caplog.at_level(logging.DEBUG):
            self.real_reg_validator.create_registries(packages=["geoips"])

        debug_logs = [
            record.message
            for record in caplog.records
            if record.levelno == logging.DEBUG
        ]
        assert any(["geoips" in message for message in debug_logs])

    def test_delete_registries_with_invalid_input(self):
        """Call delete_registries with invalid input and assert that errors are raised.

        Where the expected errors are TypeErrors or PluginRegistryErrors.
        """
        # Not a list
        with pytest.raises(TypeError):
            self.real_reg_validator.delete_registries(packages=Exception)
        # Not a list of strings
        with pytest.raises(TypeError):
            self.real_reg_validator.delete_registries(packages=[1, 2, 3, 4])
        # Not a valid package under namespace 'geoips.plugin_packages'
        with pytest.raises(PluginRegistryError):
            self.real_reg_validator.delete_registries(packages=["some_fake_package"])

    @pytest.mark.parametrize("fpath", pr_validator.registry_files, ids=generate_id)
    def test_all_registries(self, fpath):
        """Test all available yaml registries."""
        with open(fpath, "r") as fo:
            current_registry = yaml.safe_load(fo)
        self.pr_validator.validate_registry(current_registry, fpath)



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Utilities for swath compositing in GeoIPS."""

import isodate
from datetime import timedelta

from geoips.interfaces import databases


def find_preproc_alg_files(
    product_time,
    composite_window,
    sector_name,
    product,
    sensor,
    platform,
    file_format="netcdf",
    product_db=False,
    db_query_plugin=None,
    db_schemas=None,
    db_tables=None,
):
    """Find pre-processed algorithm files that were saved to disk.

    Parameters
    ----------
    product_time : datetime.datetime
        Product time
    composite_window : str
        How far back to search for pre-processed files.
        Window needs to be specified in iso8601 duration format (e.g. PT4H)
    sector_name : str
        Name of sector to composite
    product : str
        Name of product to composite
    sensor : str
        Name of sensor to composite
    platform : str
        Name of platform to composite
    file_format : str, optional
        Pre-processed file format, by default "netcdf"
    product_db : bool, optional
        Use product database to find any pre-processed file, by default False
    db_query_plugin : str, optional
        Name of product database query plugin, by default None
    db_schemas : list, optional
        Names of postgres schema to query, by default None
    db_tables : list, optional
        Names of table to query under schema, by default None

    Returns
    -------
    list
        List of pre-processed algorithm files
    """
    product_time_start = product_time - isodate.parse_duration(composite_window)
    product_time_end = product_time - timedelta(seconds=1)
    if isinstance(db_schemas, str):
        db_schemas = [db_schemas]
    if isinstance(db_tables, str):
        db_tables = [db_tables]
    if file_format == "netcdf":
        preproc_files = find_preproc_alg_netcdfs(
            product_time_start,
            product_time_end,
            sector_name,
            product,
            sensor,
            platform,
            product_db=product_db,
            postgres_query_plugin=db_query_plugin,
            postgres_schemas=db_schemas,
            postgres_tables=db_tables,
        )
    return preproc_files


def find_preproc_alg_netcdfs(
    product_time_start,
    product_time_end,
    sector_name,
    product,
    sensor,
    platform,
    product_db=False,
    postgres_query_plugin=None,
    postgres_schemas=None,
    postgres_tables=None,
):
    """Find pre-processed algorithm netCDF files that were saved to disk.

    Parameters
    ----------
    product_time_start : datetime.datetime
        Earliest product time to search for valid files
    product_time_start : datetime.datetime
     Latest product time to search for valid files
    sector_name : str
        Name of sector to composite
    product : str
        Name of product to composite
    sensor : str
        Name of sensor to composite
    platform : str
        Name of platform to composite
    file_format : str, optional
        Pre-processed file format, by default "netcdf"
    product_db : bool, optional
        Use product database to find any pre-processed file, by default False
    db_query_plugin : str, optional
        Name of product database query plugin, by default None
    db_schemas : list, optional
        Names of postgres schema to query, by default None
    db_tables : list, optional
        Names of table to query under schema, by default None

    Returns
    -------
    list
        List of pre-processed netCDF algorithm files
    """
    if product_db:
        query_plugin = databases.get_plugin(postgres_query_plugin)
        search_schema = []
        for schema in postgres_schemas:
            for ptime in [product_time_start, product_time_end]:
                schema_name = ptime.strftime(schema)
                if schema_name not in search_schema:
                    search_schema.append(schema_name)
        query_out = query_plugin(
            product_time_start,
            product_time_end,
            sector_name,
            product,
            sensor,
            platform,
            search_schema,
            file_type="nc",
            postgres_tables=postgres_tables,
        )
    nc_files = []
    for product in query_out["products"]:
        nc_file = product["productPath"]
        if nc_file not in nc_files:
            nc_files.append(nc_file)
    return sorted(nc_files)



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Data manipulation steps for model windbarbs."""
import logging
import numpy as np
import xarray as xr

LOG = logging.getLogger(__name__)

interface = "algorithms"
family = "xarray_to_numpy"
name = "model_windbarbs"


def call(
    xobj,
    output_data_range=None,
    pressure_level_range=1000,
    pressure_key=None,
    time_key="atime",
    time_fcst=-1,
    min_outbounds="crop",
    max_outbounds="crop",
    norm=False,
    inverse=False,
    time_dim=None,
    grid_geo=False,
):
    """Model-windbarbs algorithm data manipulation steps, similar to model channel.

    Parameters
    ----------
    xobj : xarray.Dataset
        * Dataset holding at minimum wind speed, direction, and pressure data
    variables : list of str
        * list of input variables used in algorithm, selects the first
    product_name : str
        * Name used to store output of algorithm to xobj

    Returns
    -------
    xarray.Dataset with new data variable named after product_name
        numpy.ndarray or numpy.MaskedArray of appropriately scaled data
    """
    # need spd and direction
    # rain flag/pressure optional

    indata = xobj["windspeed_kts"]
    indims = xobj["windspeed_kts"].dims

    # time slicing
    time_idx = indims.index(time_key)
    slice_idx = [slice(0, None)] * len(indims)

    if time_fcst == -1:
        LOG.info("Slicing for max forcast time")
        slice_idx[time_idx] = slice(time_fcst, None)
    else:
        slice_idx[time_idx] = slice(time_fcst, time_fcst + 1)

    # pressure slicing (if given)
    if pressure_key:
        pres_idx = indims.index(pressure_key)
        pressure_match = [
            p
            for p, i in enumerate(xobj["windspeed_kts"][pressure_key])
            if pressure_level_range == i
        ]
        if len(pressure_match) == 0:
            raise ValueError(
                f"No pressure found at level {pressure_level_range} for "
                + f"range of {xobj['windspeed_kts'][pressure_key]} in data"
            )

        pressure_idx = pressure_match[0]
        slice_idx[pres_idx] = slice(pressure_idx, pressure_idx + 1)

    slice_tup = tuple(i for i in slice_idx)

    spd = indata[slice_tup]
    spd = spd.squeeze()
    # should be same shape as winds
    direction = xobj["winddir"][slice_tup].squeeze()

    # rain flag
    if "rain_flag" not in list(xobj.variables):
        rain_flag = np.zeros(direction.shape)

    if output_data_range:
        from geoips.data_manipulations.corrections import apply_data_range

        spd = apply_data_range(
            spd,
            min_val=output_data_range[0],
            max_val=output_data_range[1],
            min_outbounds=min_outbounds,
            max_outbounds=max_outbounds,
            norm=norm,
            inverse=inverse,
        )

    # grid geo (lat/lon), interps expect gridded geos
    if grid_geo:
        lat, lon = np.meshgrid(
            xobj["latitude"].values, xobj["longitude"].values, indexing="ij"
        )
        xobj["latitude"] = xr.DataArray(lat, dims=("latitude", "longitude"))
        xobj["longitude"] = xr.DataArray(lon, dims=("latitude", "longitude"))

    barb_stack = np.ma.dstack((spd.values, direction.values, rain_flag)).squeeze()

    return barb_stack



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Filename specification using minimal basic attributes, and no subdirs."""

import logging
from os.path import join as pathjoin

from geoips.filenames.base_paths import PATHS as gpaths

LOG = logging.getLogger(__name__)

interface = "filename_formatters"
family = "xarray_area_product_to_filename"
name = "basic_fname"


def call(
    xarray_obj,
    area_def,
    product_name,
    output_type=".png",
    basedir=pathjoin(gpaths["ANNOTATED_IMAGERY_PATH"]),
    extra_field=None,
):
    """Create basic filename, in a specific directory (with no subdirectories).

    This filename format includes only the start time, platform name, source
    name, product name, sector name, and data provider, and a full path
    to basedir with no additional subdirectories.
    """
    sector_name = "unk"
    if area_def:
        sector_name = area_def.area_id

    path = basedir
    fname = ".".join(
        [
            xarray_obj.attrs["start_datetime"].strftime("%Y%m%d"),
            xarray_obj.attrs["start_datetime"].strftime("%H%M%S"),
            xarray_obj.attrs["platform_name"],
            xarray_obj.attrs["source_name"],
            product_name,
            sector_name,
            xarray_obj.attrs["data_provider"],
        ]
    )
    fname = "{0}.{1}".format(fname, output_type)
    return pathjoin(path, fname)



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Geoips style NetCDF output format."""
import logging

LOG = logging.getLogger(__name__)

interface = "output_formatters"
family = "xarray_data"
name = "netcdf_geoips"


def call(xarray_obj, product_names, output_fnames, clobber=False):
    """Write GeoIPS style NetCDF to disk."""
    import xarray

    prod_xarray = xarray.Dataset()

    from geoips.geoips_utils import copy_standard_metadata

    copy_standard_metadata(xarray_obj, prod_xarray)
    for product_name in product_names:
        prod_xarray[product_name] = xarray_obj[product_name]

    from geoips.plugins.modules.output_formatters.netcdf_xarray import (
        write_xarray_netcdf,
    )

    for ncdf_fname in output_fnames:
        write_xarray_netcdf(prod_xarray, ncdf_fname, clobber=clobber)
    return output_fnames



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Unit test module for testing functionality of geoips/xarray_utils/data.py."""

from importlib import resources
import logging
import pytest
from pyresample import load_area

from geoips.interfaces import sectors
from geoips.xarray_utils.data import get_minmax_latlon_from_area_def, point_in_area_def

LOG = logging.getLogger(__name__)


class XarrayDataUtils:
    """Class for testing functions out of xarray_utils/data.py.

    Currently tests get_minmax_latlon_from_area_def and point_in_area_def from the
    aforementioned file.
    """

    def __init__(self):
        """Initialize the XarrayDataUtils Class.

        Instantiate both the plugins to test and the paths to those plugins.
        """
        self._sector_plugins = self.sector_plugins
        self._sector_abspaths = self.sector_abspaths
        self._area_defs = self.area_defs

    @property
    def sector_plugins(self):
        """Available sector plugins for testing."""
        if not hasattr(self, "_sector_plugins"):
            return sectors.plugin_registry.registered_plugins["yaml_based"]["sectors"]
        else:
            return self._sector_plugins

    @property
    def sector_abspaths(self):
        """Absolute paths to the available (static) sector plugins for testing."""
        if not hasattr(self, "_sector_abspaths"):
            self._sector_abspaths = []
            for plg_name in self.sector_plugins:
                plg_family = self.sector_plugins[plg_name]["family"]
                if plg_family == "area_definition_static":
                    self._sector_abspaths.append(
                        str(
                            resources.files(self.sector_plugins[plg_name]["package"])
                            / self.sector_plugins[plg_name]["relpath"]
                        )
                    )
            return sorted(self._sector_abspaths)
        else:
            return self._sector_abspaths

    @property
    def area_defs(self):
        """Available (static) PyResample AreaDefinition()'s for testing."""
        if not hasattr(self, "_area_defs"):
            return [load_area(abspath, "spec") for abspath in self.sector_abspaths]
        else:
            return self._area_defs

    def generate_id(self, area_def):
        """Generate and Return an id for the provided PyResample AreaDefinition()."""
        return f"{area_def.area_id.title()}"


xdu = XarrayDataUtils()


@pytest.mark.parametrize("area_def", xdu.area_defs, ids=xdu.generate_id)
def test_get_minmax_latlon_from_area_def(area_def):
    """Test the function geoips/xarray_utils/data.py:get_minmax_latlon_from_area_def.

    This test will only be applied to static sectors, as dynamic sectors do not have
    enough information in the yamls to create an area definition initially.
    """
    min_lon, min_lat, max_lon, max_lat = get_minmax_latlon_from_area_def(area_def, 0, 0)
    pole = False
    if max_lat == 90:
        pole = True
        assert min_lat == list(area_def.area_extent_ll)[1]
    elif min_lat == -90:
        pole = True
        assert max_lat == list(area_def.area_extent_ll)[3]
    if pole:
        assert min_lon == -180
        assert max_lon == 180
    else:
        assert -180 <= min_lon <= max_lon < 360
        assert -90 <= min_lat <= max_lat <= 90


@pytest.mark.parametrize("area_def", xdu.area_defs, ids=xdu.generate_id)
def test_point_in_area_def(area_def):
    """Test the funciton geoips/xarray_utils/data.py:point_in_area_def.

    This test will only be applied to static sectors, as dynamic sectors do not have
    enough information in the yamls to create an area definition initially.
    """
    if area_def.area_id in ["north_pole", "south_pole"]:
        assert point_in_area_def(area_def, 0, 90) or point_in_area_def(area_def, 0, -90)
    else:
        assert not (
            point_in_area_def(area_def, 0, 90) or point_in_area_def(area_def, 0, -90)
        )



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""matplotlib geographic map (basemap or cartopy) utilities."""

# Python Standard Libraries
import logging
from copy import deepcopy
import pyresample
import cartopy.feature as cfeature
from math import ceil
import numpy as np
import matplotlib.ticker as mticker
from cartopy.mpl.gridliner import LONGITUDE_FORMATTER, LATITUDE_FORMATTER
from geoips.interfaces import feature_annotators, gridline_annotators


LOG = logging.getLogger(__name__)


def ellps2axis(ellps_name):
    """
    Get semi-major and semi-minor axis from ellipsis definition.

    Parameters
    ----------
    ellps_name : str
        Standard name of ellipsis

    Returns
    -------
    avar : float
        semi-major axis
    bvar : float
        semi-minor axis
    """
    ellps = {
        "helmert": {"a": 6378200.0, "b": 6356818.1696278909},
        "intl": {"a": 6378388.0, "b": 6356911.9461279465},
        "merit": {"a": 6378137.0, "b": 6356752.2982159676},
        "wgs72": {"a": 6378135.0, "b": 6356750.5200160937},
        "sphere": {"a": 6370997.0, "b": 6370997.0},
        "clrk66": {"a": 6378206.4000000004, "b": 6356583.7999999998},
        "nwl9d": {"a": 6378145.0, "b": 6356759.7694886839},
        "lerch": {"a": 6378139.0, "b": 6356754.2915103417},
        "evrstss": {"a": 6377298.5559999999, "b": 6356097.5503008962},
        "evrst30": {"a": 6377276.3449999997, "b": 6356075.4131402401},
        "mprts": {"a": 6397300.0, "b": 6363806.2827225132},
        "krass": {"a": 6378245.0, "b": 6356863.0187730473},
        "walbeck": {"a": 6376896.0, "b": 6355834.8466999996},
        "kaula": {"a": 6378163.0, "b": 6356776.9920869097},
        "wgs66": {"a": 6378145.0, "b": 6356759.7694886839},
        "evrst56": {"a": 6377301.2429999998, "b": 6356100.2283681016},
        "new_intl": {"a": 6378157.5, "b": 6356772.2000000002},
        "airy": {"a": 6377563.3959999997, "b": 6356256.9100000001},
        "bessel": {"a": 6377397.1550000003, "b": 6356078.9628181886},
        "seasia": {"a": 6378155.0, "b": 6356773.3205000004},
        "aust_sa": {"a": 6378160.0, "b": 6356774.7191953054},
        "wgs84": {"a": 6378137.0, "b": 6356752.3142451793},
        "hough": {"a": 6378270.0, "b": 6356794.3434343431},
        "wgs60": {"a": 6378165.0, "b": 6356783.2869594367},
        "engelis": {"a": 6378136.0499999998, "b": 6356751.3227215428},
        "apl4.9": {"a": 6378137.0, "b": 6356751.796311819},
        "andrae": {"a": 6377104.4299999997, "b": 6355847.4152333336},
        "sgs85": {"a": 6378136.0, "b": 6356751.301568781},
        "delmbr": {"a": 6376428.0, "b": 6355957.9261637237},
        "fschr60m": {"a": 6378155.0, "b": 6356773.3204827355},
        "iau76": {"a": 6378140.0, "b": 6356755.2881575283},
        "plessis": {"a": 6376523.0, "b": 6355863.0},
        "cpm": {"a": 6375738.7000000002, "b": 6356666.221912113},
        "fschr68": {"a": 6378150.0, "b": 6356768.3372443849},
        "mod_airy": {"a": 6377340.1890000002, "b": 6356034.4460000005},
        "grs80": {"a": 6378137.0, "b": 6356752.3141403561},
        "bess_nam": {"a": 6377483.8650000002, "b": 6356165.3829663256},
        "fschr60": {"a": 6378166.0, "b": 6356784.2836071067},
        "clrk80": {"a": 6378249.1449999996, "b": 6356514.9658284895},
        "evrst69": {"a": 6377295.6639999999, "b": 6356094.6679152036},
        "grs67": {"a": 6378160.0, "b": 6356774.5160907144},
        "evrst48": {"a": 6377304.0630000001, "b": 6356103.0389931547},
    }
    try:
        ellps_axis = ellps[ellps_name.lower()]
        avar = ellps_axis["a"]
        bvar = ellps_axis["b"]
    except KeyError:
        raise ValueError(
            (
                "Could not determine semi-major and semi-minor axis "
                "of specified ellipsis %s"
            )
            % ellps_name
        )
    return avar, bvar


def is_crs(mapobj):
    """
    Determine if the map object we are using is a cartopy crs instance.

    Parameters
    ----------
    mapobj : map object
        Basemap or cartopy._PROJ4Projection object

    Returns
    -------
    crs : bool
        True if it is a CRS object, False otherwise.
    """
    crs = False
    if "cartopy" in str(mapobj.__class__):
        crs = True
    return crs


def parallels(area_def, grid_size):
    """
    Calculate the parallels (latitude) that fall within the input sector.

    Parameters
    ----------
    area_def : AreaDefinition
        pyresample AreaDefinition
    grid_size : float
        grid spacing in degrees

    Returns
    -------
    lat_ticks : list
        latitude locations for gridlines
    """
    if grid_size == "auto":
        grid_size = compute_lat_auto_spacing(area_def)

    gs = float(grid_size)

    lats = area_def.get_lonlats()[1]
    mlats = np.ma.masked_greater(lats, 90)
    min_parallel = ceil(float(mlats.min()) / gs) * gs
    max_parallel = ceil(float(mlats.max()) / gs) * gs
    lat_ticks = np.arange(min_parallel, max_parallel, gs)
    LOG.info("List of parallels: min/max %s, %s", lat_ticks[0], lat_ticks[-1])
    lat_ticks = [round(lat_tick, 1) for lat_tick in lat_ticks]
    return lat_ticks


def meridians(area_def, grid_size):
    """
    Calculate the meridians (longitude) that are within the input sector.

    Parameters
    ----------
    area_def : AreaDefinition
        pyresample AreaDefinition
    grid_size : float
        grid spacing in degrees

    Returns
    -------
    meridians_to_draw : list
        longitude locations for gridlines
    """
    if grid_size == "auto":
        grid_size = compute_lon_auto_spacing(area_def)

    gs = float(grid_size)

    corners = area_def.corners
    lons = [np.rad2deg(corn.lon) for corn in corners]
    llcrnrlon = lons[3]
    urcrnrlon = lons[1]

    # Needed for full disk - need to generalize so it works for both.
    # mlons = np.ma.masked_greater(sector.area_definition.get_lonlats()[0],180)
    # corners = mlons.min(),mlons.max()
    # #lons = [np.rad2deg(corn.lon) for corn in corners]
    # llcrnrlon = corners[0]
    # urcrnrlon = corners[1]

    cent_lon = area_def.proj_dict["lon_0"]
    if urcrnrlon < cent_lon < llcrnrlon:
        urcrnrlon += 360
    elif urcrnrlon < llcrnrlon:
        llcrnrlon -= 360

    min_meridian = ceil(float(llcrnrlon) / gs) * gs
    max_meridian = ceil(float(urcrnrlon) / gs) * gs
    meridians_to_draw = np.arange(min_meridian, max_meridian, gs)

    meridians_to_draw = pyresample.utils.wrap_longitudes(meridians_to_draw)
    LOG.info(
        "List of meridians: min/max %s, %s", meridians_to_draw[0], meridians_to_draw[-1]
    )
    meridians_to_draw = [round(lon_tick, 1) for lon_tick in meridians_to_draw]
    return meridians_to_draw


def check_gridline_annotator(gridline_annotator):
    """
    Check gridlines_info dictionary for that all required fields.

    Parameters
    ----------
    gridline_annotator : YamlPlugin
        A gridline annotator plugin instance.

    Raises
    ------
    ValueError
        If required field is missing
    """
    required_fields = {
        "spacing": ["latitude", "longitude"],
        "labels": ["top", "bottom", "left", "right"],
        "lines": ["color", "linestyle", "linewidth"],
    }

    if gridline_annotator is None:
        gridline_annotator = gridline_annotators.get_plugin("default")

    for key, subkeys in required_fields.items():
        if key not in gridline_annotator["spec"]:
            raise ValueError(
                "Missing gridlines_info entry {0}, required_fields {1}".format(
                    key, required_fields
                )
            )
        if key in ["xpadding" or "ypadding"]:
            # We skip here since we don't require 'xpadding' or 'ypadding'. These are
            # optional attributes that you can apply to the gridline_annotator if
            # wanted.
            continue
        for subkey in subkeys:
            if subkey not in gridline_annotator["spec"][key]:
                LOG.info(gridline_annotator)
                raise ValueError(
                    "Missing gridline_annotator property "
                    f"{key}.{subkey}, required_fields {required_fields}"
                )
    return gridline_annotator


def set_gridlines_info_dict(gridlines_info, area_def):
    """
    Set plotting gridlines.

    Set the final values for gridlines plotting params, pulling from
    argument and defaults.

    Parameters
    ----------
    gridlines_info : dict
        Dictionary of parameters for plotting gridlines. The following
        fields are available.  If a field is not included in the
        dictionary, the field is added to the return dictionary and the
        default is used.
    area_def : AreaDefinition
        pyresample AreaDefinition

    Returns
    -------
    use_gridlines_info : dict
        gridlines_info dictionary, with fields as specified above.

    Notes
    -----
    Defaults specified as::

        gridlines_info['grid_lat_spacing']      default auto calculated 5 lat grid lines
        gridlines_info['grid_lon_spacing']      default auto calculated 5 lon grid lines
        gridlines_info['grid_lat_xoffset']      default None (0.01 * image height)
        gridlines_info['grid_lon_xoffset']      default None (0.01 * image width)
        gridlines_info['grid_lat_yoffset']      default None (0.01 * image height)
        gridlines_info['grid_lon_yoffset']      default None (0.01 * image width)
        gridlines_info['grid_lat_fontsize']     default None (plot fontsize)
        gridlines_info['grid_lon_fontsize']     default None (plot fontsize)
        gridlines_info['left_label']            default True
        gridlines_info['right_label']           default False
        gridlines_info['top_label']             default True
        gridlines_info['bottom_label']          default False
        gridlines_info['grid_lat_linewidth']    default 1
        gridlines_info['grid_lon_linewidth']    default 1
        gridlines_info['grid_lat_color']        default 'black'
        gridlines_info['grid_lon_color']        default 'black'
        gridlines_info['grid_lat_dashes']       default [4, 2]
        gridlines_info['grid_lon_dashes']       default [4, 2]
    """
    use_gridlines_info = {}

    if (
        gridlines_info is None
        or "grid_lat_spacing" not in gridlines_info.keys()
        or "grid_lon_spacing" not in gridlines_info.keys()
        or gridlines_info["grid_lon_spacing"] == "auto"
        or gridlines_info["grid_lat_spacing"] == "auto"
    ):
        minlat = area_def.area_extent_ll[1]
        maxlat = area_def.area_extent_ll[3]
        minlon = pyresample.utils.wrap_longitudes(area_def.area_extent_ll[0])
        maxlon = pyresample.utils.wrap_longitudes(area_def.area_extent_ll[2])
        if minlon > maxlon and maxlon < 0:
            maxlon = maxlon + 360
        lon_extent = maxlon - minlon
        lat_extent = maxlat - minlat

        if lon_extent > 5:
            lon_spacing = int(lon_extent / 5)
        elif lon_extent > 2.5:
            lon_spacing = 1
        else:
            lon_spacing = lon_extent / 5.0

        if lat_extent > 5:
            lat_spacing = int(lat_extent / 5)
        elif lat_extent > 2.5:
            lat_spacing = 1
        else:
            lat_spacing = lat_extent / 5.0

        LOG.info("lon_extent: %s, lon_spacing: %s", lon_extent, lon_spacing)
        use_gridlines_info["grid_lat_spacing"] = lat_spacing
        use_gridlines_info["grid_lon_spacing"] = lon_spacing

    use_gridlines_info["left_label"] = True
    use_gridlines_info["right_label"] = False
    use_gridlines_info["top_label"] = True
    use_gridlines_info["bottom_label"] = False
    use_gridlines_info["grid_lat_linewidth"] = 1
    use_gridlines_info["grid_lon_linewidth"] = 1
    use_gridlines_info["grid_lat_color"] = "black"
    use_gridlines_info["grid_lon_color"] = "black"
    use_gridlines_info["grid_lat_xoffset"] = None
    use_gridlines_info["grid_lon_xoffset"] = None
    use_gridlines_info["grid_lat_yoffset"] = None
    use_gridlines_info["grid_lon_yoffset"] = None
    use_gridlines_info["grid_lat_fontsize"] = None
    use_gridlines_info["grid_lon_fontsize"] = None
    use_gridlines_info["grid_lat_dashes"] = [4, 2]
    use_gridlines_info["grid_lon_dashes"] = [4, 2]

    if gridlines_info is not None:
        for gkey in gridlines_info.keys():
            if gridlines_info[gkey] is not None:
                use_gridlines_info[gkey] = gridlines_info[gkey]

    return use_gridlines_info


def compute_lat_auto_spacing(area_def):
    """Compute automatic spacing for latitude lines based on area definition."""
    minlat = area_def.area_extent_ll[1]
    maxlat = area_def.area_extent_ll[3]
    lat_extent = maxlat - minlat
    if lat_extent > 5:
        lat_spacing = int(lat_extent / 5)
    elif lat_extent > 2.5:
        lat_spacing = 1
    elif lat_extent == 0:
        lat_spacing = 2
    else:
        lat_spacing = lat_extent / 5.0
    return lat_spacing


def compute_lon_auto_spacing(area_def):
    """Compute automatic spacing for longitude lines based on area definition."""
    minlon = pyresample.utils.wrap_longitudes(area_def.area_extent_ll[0])
    maxlon = pyresample.utils.wrap_longitudes(area_def.area_extent_ll[2])
    if minlon > maxlon and maxlon < 0:
        maxlon = maxlon + 360
    lon_extent = maxlon - minlon

    if lon_extent > 5:
        lon_spacing = int(lon_extent / 5)
    elif lon_extent > 2.5:
        lon_spacing = 1
    else:
        lon_spacing = lon_extent / 5.0
    return lon_spacing


def check_feature_annotator(feature_annotator):
    """Check that the provided feature_annotator plugin has all required fields.

    Parameters
    ----------
    feature_annotator : YamlPlugin
        A feature annotator plugin.

    Raises
    ------
    ValueError
        if any field is missing

    See Also
    --------
    geoips.image_utils.maps.get_feature_annotator
        For complete list of fields, and appropriate defaults
    """
    if feature_annotator is None:
        feature_annotator = feature_annotators.get_plugin("default")
    spec = feature_annotator["spec"]

    feature_types = ["coastline", "borders", "rivers", "states"]
    for feature_name in feature_types:
        if feature_name not in spec:
            raise ValueError(
                f"Missing '{feature_name}' property in feature_annotator "
                f"named '{feature_annotator['name']}'"
            )
        if "enabled" not in spec[feature_name]:
            raise ValueError(
                f"Missing 'enabled' property of '{feature_name}' in feature_annotator "
                f"named '{feature_annotator['name']}'"
            )
        if spec[feature_name]["enabled"]:
            props = ["edgecolor", "linewidth"]
            for prop in props:
                if prop not in spec[feature_name]:
                    raise ValueError(
                        f"Missing '{prop}' property of "
                        f"'{feature_name}' in feature_annotator, "
                        f"named '{feature_annotator['name']}'."
                    )
    return feature_annotator


# def set_boundaries_info_dict(boundaries_info):
#     """Set the boundary information.
#
#     Set the final values for coastlines, states, countries plotting params,
#     pulling from argument and defaults.
#
#     Parameters
#     ----------
#     boundaries_info : dict
#         Dictionary of parameters for plotting gridlines.
#         The following fields are available.  If a field is not included in the
#         dictionary, the field is added to the return dictionary and the default
#         is used (see defaults in Notes below).
#
#     Returns
#     -------
#     use_boundaries_info : dict
#         boundaries_info dictionary, with fields as specified above.
#
#     Notes
#     -----
#     Defaults specified as::
#
#         boundaries_info['request_coastlines']       default True
#         boundaries_info['request_countries']        default True
#         boundaries_info['request_states']           default True
#         boundaries_info['request_rivers']           default True
#
#         boundaries_info['coastlines_linewidth']     default 2
#         boundaries_info['countries_linewidth']      default 1
#         boundaries_info['states_linewidth']         default 0.5
#         boundaries_info['rivers_linewidth']         default 0
#
#         boundaries_info['coastlines_color']         default 'red'
#         boundaries_info['countries_color']          default 'red'
#         boundaries_info['states_color']             default 'red'
#         boundaries_info['rivers_color']             default 'red'
#     """
#     use_boundaries_info = {}
#     use_boundaries_info["request_coastlines"] = True
#     use_boundaries_info["request_countries"] = True
#     use_boundaries_info["request_states"] = True
#     use_boundaries_info["request_rivers"] = True
#
#     use_boundaries_info["coastlines_linewidth"] = 2
#     use_boundaries_info["countries_linewidth"] = 1
#     use_boundaries_info["states_linewidth"] = 0.5
#     use_boundaries_info["rivers_linewidth"] = 0
#
#     use_boundaries_info["coastlines_color"] = "red"
#     use_boundaries_info["countries_color"] = "red"
#     use_boundaries_info["states_color"] = "red"
#     use_boundaries_info["rivers_color"] = "red"
#
#     # Grab any values that were passed
#     if boundaries_info is not None:
#         for bkey in boundaries_info.keys():
#             if boundaries_info[bkey] is not None:
#                 use_boundaries_info[bkey] = boundaries_info[bkey]
#
#     return use_boundaries_info


def draw_features(mapobj, curr_ax, feature_annotator, zorder=None):
    """Draw cartopy features.

    Draw features on specified cartopy map instance, based
    on specs found in the feature_annotator plugin.

    Parameters
    ----------
    mapobj : map object
        CRS object for plotting map features
    curr_ax : matplotlib.axes._axes.Axes
        matplotlib Axes object for plotting map features
    feature_annotator : dict
        Dictionary of parameters for plotting map features
    zorder : int, optional
        The matplotlib zorder

    See Also
    --------
    geoips.image_utils.maps.check_feature_annotator
          for required dictionary entries and defaults.
    """
    LOG.info("Drawing coastlines, countries, states, rivers")
    feature_annotator = check_feature_annotator(feature_annotator)

    # NOTE passing zorder=None is NOT the same as not passing zorder.
    # Ensure if zorder is passed, it is propagated to cartopy.
    extra_args = {}
    if zorder is not None:
        extra_args = {"zorder": zorder}

    LOG.info("    Plotting with cartopy")

    for name, feature in feature_annotator["spec"].items():
        feat = deepcopy(feature)
        # Need to make sure this isn't 'background' as this isn't a feature that can be
        # added to an axis. This is just the background color of where those features
        # will or will not be added
        if name != "background" and feat.pop("enabled"):
            curr_ax.add_feature(getattr(cfeature, name.upper()), **feat, **extra_args)


def draw_gridlines(mapobj, area_def, curr_ax, gridline_annotator, zorder=None):
    """Draw gridlines on map object.

    Draw gridlines on a cartopy map object, as specified by a
    gridline_annotator plugin instance

    Parameters
    ----------
    mapobj : map object
        CRS object for plotting gridlines
    area_def : AreaDefinition
        pyresample AreaDefinition object
    curr_ax : matplotlib.axes._axes.Axes
        matplotlib Axes object for plotting gridlines
    gridline_annotator : YamlPlugin
        A gridline_annotator plugin instance
    zorder : int, optional
        The matplotlib zorder value

    See Also
    --------
    geoips.image_utils.maps.get_gridlines_info_dict
        For complete list of fields, and appropriate default
    """
    LOG.info("Drawing gridlines")
    gridline_annotator = check_gridline_annotator(gridline_annotator)

    # NOTE passing zorder=None is NOT the same as not passing zorder.
    # Ensure if zorder is passed, it is propagated to cartopy.
    extra_args = {}
    if zorder is not None:
        extra_args = {"zorder": zorder}

    LOG.info("    Plotting with cartopy")

    spec = gridline_annotator["spec"]
    labels = spec["labels"]

    if "xpadding" in labels.keys():
        extra_args["xpadding"] = labels["xpadding"]
    if "ypadding" in labels.keys():
        extra_args["ypadding"] = labels["ypadding"]

    # Note: linestyle is specified by a tuple: (offset, (pixels_on,
    # pixels_off, pxon, pxoff)), etc
    lines = deepcopy(spec["lines"])
    lines["linestyle"] = (0, tuple(lines["linestyle"]))
    glnr = curr_ax.gridlines(draw_labels=True, **lines, **extra_args)

    # Default to False, unless set in argument
    glnr.top_labels = labels["top"]
    glnr.bottom_labels = labels["bottom"]
    glnr.left_labels = labels["left"]
    glnr.right_labels = labels["right"]

    spacing = spec["spacing"]
    lat_ticks = parallels(area_def, spacing["latitude"])
    lon_ticks = meridians(area_def, spacing["longitude"])

    LOG.debug(f"Lon Ticks: {lon_ticks}")
    glnr.xlocator = mticker.FixedLocator(lon_ticks)
    LOG.debug(f"Lat Ticks: {lat_ticks}")
    glnr.ylocator = mticker.FixedLocator(lat_ticks)

    glnr.xformatter = LONGITUDE_FORMATTER
    glnr.yformatter = LATITUDE_FORMATTER
    glnr.xlabel_style = {"rotation": 0}
    glnr.ylabel_style = {"rotation": 0}

    for style_arg in gridline_annotators.label_kwargs.keys():
        # If a (x/y)label_style kwarg is present in the gridline annotator's label
        # object, than make sure to apply that argument now. Ie. color of the label,
        # it's offset, etc. See geoips.schema.gridline_annotators.cartopy:labels for
        # more information.
        if style_arg in labels.keys():
            glnr.xlabel_style[style_arg] = labels[style_arg]
            glnr.ylabel_style[style_arg] = labels[style_arg]



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Semi-Abstract CLI Test Class implementing attributes shared by commands."""

import abc
import contextlib
from importlib import metadata
import io
from numpy import any
import pytest
import subprocess
import sys

from geoips.commandline.cmd_instructions import alias_mapping
from geoips.commandline.commandline_interface import GeoipsCLI
from geoips.commandline.commandline_interface import main as cli_main
from geoips.geoips_utils import is_editable


gcli = GeoipsCLI()


def generate_id(args):
    """Generate an ID from the arguments provided.

    Differs slightly from BaseCliTest:generate_id as this function might encounter
    arguments that are a dictionary. This happens due to how those tests are structured,
    where two sets of args are provided for one test (cli_args, expected), where
    expected is a dictionary and should not be part of the test ID.

    Parameters
    ----------
    args: list[str]
        - A list of strings representing the CLI command being ran.
    """
    if isinstance(args, dict):
        return ""
    return " ".join(args)


def check_valid_command_using_monkeypatch(monkeypatch, cli_args, expected):
    """Assert that the command provided works as expected.

    This method is used in multiple unit test calls that don't make use of BaseCliTest.
    Essentially, we mock a sys.argv call via monkeypatch and assert that the output
    of the command casted as vars(args) matches a dictionary of expected arguments that
    the test provides.

    Parameters
    ----------
    monkeypatch: Generator[Monkeypatch]
        - A pytest fixture used to create isolated and controlled test environments by
          modifying code at runtime.
    cli_args: list[str]
        - A list of strings representing the CLI command being ran.
    expected: dict
        - A dictionary of expected argument values that are returned from the command
          being ran from args.
    """
    monkeypatch.setattr("sys.argv", cli_args)
    args = cli_main(suppress_args=False)
    arg_dict = vars(args)
    arg_dict.pop("command_parser")
    arg_dict.pop("exe_command")
    print(arg_dict)
    # Assert that the argument dictionary returned from the CLI matches what we expect
    assert arg_dict == expected


class BaseCliTest(abc.ABC):
    """Top-Level CLI Test Class which implements shared attributes for commands."""

    alias_mapping = alias_mapping

    def generate_id(self, args):
        """Generate an ID for the test-arguments provided."""
        return " ".join(args)

    @property
    def plugin_package_names(self):
        """List of names of every installed GeoIPS package."""
        if not hasattr(self, "_plugin_package_names"):
            self._plugin_package_names = [
                ep.value for ep in metadata.entry_points(group="geoips.plugin_packages")
            ]
        return self._plugin_package_names

    @property
    def test_datasets(self):
        """List of every available GeoIPS test dataset name."""
        if not hasattr(self, "_test_datasets"):
            self._test_datasets = [
                "test_data_amsr2",
                "test_data_clavrx",
                "test_data_fusion",
                "test_data_gpm",
                "test_data_noaa_aws",
                "test_data_sar",
                "test_data_scat",
                "test_data_smap",
                "test_data_viirs",
            ]
        return self._test_datasets

    def retrieve_selected_columns(self, args):
        """If --columns was used for a command, retrieve the list of columns.

        Only used for commands that invoke --columns as a valid optional argument.
        Currently, only a '--columns' option exists for list commands, and this function
        will retrieve the selected list of columns so we can assert that they, and
        nothing else, are listed correctly in the output of the list command.

        We expect to add the --columns argument to the 'get' command in the future.

        Parameters
        ----------
        args: list of strings
            - Essentially <command_sent_to_cli>.split(" ")
            - ie. "geoips list interface algorithms --columns package interface relpath"
              would become:
              ["geoips", "list", "interface", "algorithms",
              "--columns", "package", "interface", "relpath"]
        """
        selected_cols = None
        if "--columns" in args:
            start_idx = None
            for idx, arg in enumerate(args):
                if arg == "--columns":
                    start_idx = idx + 1
                    break
            if start_idx is None or start_idx >= len(args):
                usage_str = f"{' '.join(args[:3])} -h for more information."
                pytest.fail(
                    f"Unexpected usage of --columns arg. Please run {usage_str}"
                )
            selected_cols = args[start_idx:]
        return selected_cols

    def assert_correct_headers_in_output(self, output, headers, selected_cols):
        """Ensure that all selected columns are in the corresponding 'list' output.

        This is only applied to list commands, as they produce tabular output with
        headers. Given output, a list of headers, and selected columns, check that
        each corresponding header to each selected column is in the provided output.

        Parameters
        ----------
        output: multiline str
            - The caputured output of the terminal
        headers: dict of strings
            - The Corresponding Header, column_val dict to parse through
        selected_columns: list of strings or None
            - If None, asser that every Header key is in the output. Otherwise check
              that the corresponding Header key, column_val is in the output
        """
        for header in headers:
            if selected_cols is None or headers[header] in selected_cols:
                assert header in output or "has no" in output or "No plugins" in output

    def assert_non_editable_error_or_wrong_package(self, args, error):
        """If we found a package in non-editable mode, assert that an error exists.

        This is used for commands which require packages to be installed in editable
        mode to work (list scripts, test scripts, list unit-tests, test-linting). Since
        we'll check for the same errors in all of these tests, we've refactored it to
        be in one location.

        Parameters
        ----------
        args: list of str
            - The arguments provided to the CLI
        error: str
            - The error output of the CLI call

        Returns
        -------
        editable: bool
            - The truth value of whether or not all packages were editable.
        """
        # Test if a package was specified and is installed in non-editable mode
        pkg_idx = -1
        pkg_name = None
        # Default to last item of args. If no argument is provided after -p, an
        # error will be raised later anyways.
        for idx, arg in enumerate(args):
            if arg == "-p" and len(args) > idx + 1:
                # That means a package was provided. It might not be valid, but we
                # don't care. We'll test this in the if statements below.
                pkg_idx = idx + 1
                pkg_name = args[pkg_idx]
                break

        if pkg_name is not None and pkg_name not in self.plugin_package_names:
            # If the package provided is not a valid package, check for that error
            # instead
            assert f"{args[2]}: error: argument --package_name/-p: invalid" in error
            return False

        if pkg_name:
            # Just test if this package is in editable mode
            editable = is_editable(pkg_name)
        else:
            # Otherwise, assume we're working on all installed packages
            editable = any(
                [is_editable(pkg_name) for pkg_name in self.plugin_package_names]
            )

        if not editable:
            if "-p" in args and "--integration" in args and "geoips" not in args[1:]:
                # This is a specific case for the integration test scripts that
                # only work for geoips. Make sure an error is raised that says
                # we cannot run integration tests in packages other than 'geoips'
                integration_error = (
                    "script: error: Only package 'geoips' has integration tests"
                )
                package_name_error = (
                    "error: argument --package_name/-p: invalid choice:"
                )
                assert integration_error in error or package_name_error in error
            else:
                # If the package provided is a valid installed package, assert that
                # an non-editable error was raised
                assert "is installed in non-editable mode" in error
        return editable

    @property
    @abc.abstractmethod
    def command_combinations(self):
        """A 2d List of command combinations for a CLI command call.

        This can cover every possible combination to call a command or it can be a
        subset of commands if they take a while to run.

        For a shorter command such as 'geoips list interface', this property would be
        every possible combination of strings used to call 'geoips list interface'.
        This would take the form of:

        - [
            - ["geoips", "list", "interface", "algorithms", "-p", "data_fusion"],
            - ["geoips", "list", "interface", "algorithms", "-p", "geoips"],
            - ["geoips", "list", "interface", "algorithms", "-p", "geoips_clavrx"],
            - ...
            - ["geoips", "list", "interface", <interface_name>, "-p", <pkg_name>],
            - ["geoips", "list", "interface", <interface_name>],
            - ["geoips", "list", "interface", <invalid_interface_name>],
            - ["geoips", "list", "interface", <interface_name>, "-p", <bad_pkg_name>],
            - ...
        - ]

        For a longer command such as 'geoips test script', we grab a stochastic subset
        of the possible commands and use those to test instead.
        """
        pass

    @abc.abstractmethod
    def check_error(self, args, error):
        """Ensure that the 'geoips <cmd> ...' error output is correct.

        Parameters
        ----------
        args: 2D list of str
            - The arguments used to call the CLI (expected to fail)
        error: str
            - Multiline str representing the error output of the CLI call
        """
        pass

    @abc.abstractmethod
    def check_output(self, args, output):
        """Ensure that the 'geoips <cmd> ...' successful output is correct.

        Parameters
        ----------
        args: 2D list of str
            - The arguments used to call the CLI
        output: str
            - Multiline str representing the output of the CLI call
        """
        pass

    def capture_output(self, func, args):
        """Redirect stdout and stderr output from a python function to two variables.

        Where these two variables (stdout, stderr), will be used for testing the output
        of CLI commands to make sure they're working as expected.

        Parameters
        ----------
        func: python function
            - The function whose output we want to capture.
        args: array of str
            - List of arguments to call the CLI with (ie. ['geoips', '<cmd>'])

        Returns
        -------
        stdout: str
            - The ouput that would have been printed to sys.stdout.
        stderr: str
            - The output that would have been printed to sys.stderr.
        """
        stdout_capture = io.StringIO()
        stderr_capture = io.StringIO()

        with (
            contextlib.redirect_stdout(stdout_capture),
            contextlib.redirect_stderr(stderr_capture),
        ):
            try:
                func()
            except SystemExit as se:
                print(se, file=sys.stderr)
            except Exception as e:
                print(e, file=sys.stderr)
            # except RuntimeError as re:
            #     print(re, file=sys.stderr)

        # Retrieve the output printed from 'func'
        stdout = stdout_capture.getvalue()
        stderr = stderr_capture.getvalue()

        # Close the StringIO Objects
        stdout_capture.close()
        stderr_capture.close()

        return stdout, stderr

    def viable_monkeypatch(self, args):
        """Determine whether or not the arguments are viable to run via monkeypatch.

        We are currently unable to catch certain commands outputs using
        contextlib.redirect_stdout / contextlib.redirect_stderr. The cause of this is
        still unknown, however we have a workaround using subprocess for the time being.

        If the arguments can be ran via monkeypatch, return True, otherwise return
        False.

        Parameters
        ----------
        args: array of str
            - List of arguments to call the CLI with (ie. ['geoips', '<cmd>'])

        Returns
        -------
        monkeypatch_viable: bool
            - Whether or not the arguments can be ran with monkeypatch. Monkeypatch is a
              much quicker option if available.
        """
        match args:
            case _ if "-h" in args:
                # Can't capture help messages using monkeypatch... yet
                return False
            case _ if "linting" in args:
                # Can't capture linting output using monkeypatch... yet
                return False
            case _ if "test" in args and "script" in args:
                # Can't capture bash script output using monkeypatch... yet
                return False
            case _ if (
                "run" in args
                or "run_procflow" in args
                or "data_fusion_procflow" in args
            ):
                # Can't capture procflow output using monkeypatch... yet
                return False
            case _ if any(["non_existent" in arg for arg in args]):
                # Can't capture argparse.ArgumentError output using monkeypatch... yet
                return False
            case _ if "--long" in args and "--columns" in args:
                # Can't capture argparse.ArgumentError output using monkeypatch... yet
                return False
            case _ if "--max-depth" in args and "-1" in args:
                # Can't capture argparse.ArgumentError output using monkeypatch... yet
                return False
            case _:
                # Monkeypatch works for the provided arguments!
                return True

    def test_command_combinations(self, monkeypatch, args=None):
        """Test all or a stochastic subset of 'geoips <cmd> ...' command combinations.

        This test covers a stochastic or complete list of command combinations for all
        'geoips' commands. We also test invalid commands, to ensure that the proper help
        documentation is provided for those using the command incorrectly.

        Parameters
        ----------
        args: array of str
            - List of arguments to call the CLI with (ie. ['geoips', '<cmd>'])
        """
        if args is None:
            return
        monkeypatch_viable = self.viable_monkeypatch(args)
        if monkeypatch_viable:
            # The arguments provided were valid for monkeypatch so we will be using it
            # to execute this test
            orig_argv = sys.argv
            monkeypatch.setattr(sys, "argv", [sys.argv[0]] + args[1:])
            # Capture the output of 'execute_command'
            output, error = self.capture_output(gcli.execute_command, args)
            # Reset sys.argv to what it was originally
            sys.argv = orig_argv
        else:
            # Arguments provided were not valid for monkeypatch and we will be capturing
            # output via subprocess piping
            # Call the CLI via the provided commands with subprocess.Popen
            prc = subprocess.Popen(
                args,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            # Capture the output using subprocess.PIPE, then decode it.
            output, error = prc.communicate()
            output, error = output.decode(), error.decode()
            prc.terminate()
        assert len(output) or len(error)  # assert that some output was created
        if len(error) and (not len(output) or output == "\n"):
            self.check_error(args, error)
        else:
            print(output)
            self.check_output(args, output)



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Test script for representative product comparisons."""

import subprocess
import logging
from os.path import splitext

from geoips.commandline.log_setup import log_with_emphasis

LOG = logging.getLogger(__name__)

interface = "output_checkers"
family = "standard"
name = "geotiff"


def get_test_files_long(test_data_dir):
    """Return a Series of GeoTIFF paths, randomly modified from compare."""
    import rasterio
    import shutil
    from os import makedirs
    from os.path import exists, join
    from importlib.resources import files
    import numpy as np

    savedir = join(test_data_dir, "scratch", "unit_tests", "test_geotiffs")
    if not exists(savedir):
        makedirs(savedir)

    tif_name = "20200405_000000_SH252020_ahi_himawari-8_WV_100kts_100p00_1p0.tif"
    tif_path = str(files("geoips") / "../tests/outputs/ahi.tc.WV.geotiff" / tif_name)
    shutil.copy(tif_path, join(savedir, "compare.tif"))
    # Prepare paths for matched, close_mismatch, and bad_mismatch
    compare_file = join(savedir, "compare.tif")
    matched_file = join(savedir, "matched.tif")
    close_mismatch_file = join(savedir, "close_mismatch.tif")
    bad_mismatch_file = join(savedir, "bad_mismatch.tif")
    # Load the original GeoTIFF file
    with rasterio.open(compare_file) as src:
        compare_data = src.read()
        profile = src.profile

    # Save the original as 'compare'
    with rasterio.open(compare_file, "w", **profile) as dst:
        dst.write(compare_data)

    # Make a 'matched' version (identical to compare)
    with rasterio.open(matched_file, "w", **profile) as dst:
        dst.write(compare_data)

    # Make a 'close_mismatch' version (slightly modified)
    close_mismatch_data = compare_data + np.random.normal(
        scale=0.05, size=compare_data.shape
    )
    with rasterio.open(close_mismatch_file, "w", **profile) as dst:
        dst.write(close_mismatch_data)

    # Make a 'bad_mismatch' version (strongly modified)
    bad_mismatch_data = compare_data + np.random.normal(
        scale=0.25, size=compare_data.shape
    )
    with rasterio.open(bad_mismatch_file, "w", **profile) as dst:
        dst.write(bad_mismatch_data)

    return compare_file, [matched_file, close_mismatch_file, bad_mismatch_file]


def perform_test_comparisons_long(plugin, compare_file, test_files):
    """Test the comparison of two GeoTIFF files with the GeoTIFF Output Checker."""
    for path_idx in range(len(test_files)):
        retval = plugin.module.outputs_match(
            plugin,
            test_files[path_idx],
            compare_file,
        )
        if path_idx == 0:
            assert retval is True
        else:
            assert retval is False


def correct_file_format(fname):
    """Determine if fname is a geotiff file.

    Parameters
    ----------
    fname : str
        Name of file to check.

    Returns
    -------
    bool
        True if it is a geotiff file, False otherwise.
    """
    if splitext(fname)[-1] in [".tif"]:
        return True
    return False


def outputs_match(plugin, output_product, compare_product):
    """Use diff system command to compare currently produced image to correct image.

    Parameters
    ----------
    plugin: OutputCheckerPlugin
        The correspdonding geotiff output_checker - not used but needed in signature
    output_product : str
        Full path to current output product
    compare_product : str
        Full path to comparison product

    Returns
    -------
    bool
        Return True if images match, False if they differ
    """
    # out_diffimg = get_out_diff_fname(compare_product, output_product)

    call_list = ["diff", output_product, compare_product]
    LOG.info("Running %s", " ".join(call_list))
    retval = subprocess.call(call_list)

    # subimg_retval = subprocess.call(call_list)
    if retval != 0:
        log_with_emphasis(
            LOG.interactive,
            "BAD geotiffs do NOT match exactly",
            f"output_product: {output_product}",
            f"compare_product: {compare_product}",
        )
        return False

    log_with_emphasis(LOG.info, "GOOD geotiffs match")
    return True


def call(plugin, compare_path, output_products):
    """Compare the "correct" geotiffs found the list of current output_products.

    Compares files produced in the current processing run with the list of
    "correct" files contained in "compare_path".

    Parameters
    ----------
    plugin: OutputCheckerPlugin
        The corresponding geotiff OutputCheckerPlugin that has access to needed methods
    compare_path : str
        Path to directory of "correct" products - filenames must match output_products
    output_products : list of str
        List of strings of current output products,
        to compare with products in compare_path

    Returns
    -------
    int
        Binary code: 0 if all comparisons were completed successfully.
    """
    retval = plugin.compare_outputs(
        compare_path,
        output_products,
    )
    return retval



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""VIIRS SDR Satpy reader.

This VIIRS reader is designed for reading the NPP/JPSS SDR HDF5 files.
The input files are produced by CSPP Polar (CSPP RDR pipeline),
and the read by satpy.

V1.1.0:  NRL-Monterey, Aug. 2024

"""

# Python Standard Libraries
import logging
import os

# Third-Party Libraries
import h5py
import numpy as np
from pandas import DataFrame, date_range, to_datetime
from scipy.interpolate import NearestNDInterpolator
import xarray as xr

# GeoIPS Libraries
from geoips.plugins.modules.readers.utils.geostationary_geolocation import get_indexes
from geoips.utils.context_managers import import_optional_dependencies

# If this reader is not installed on the system, don't fail altogether, just skip this
# import. This reader will not work if the import fails, and the package will have to be
# installed to process data of this type.

LOG = logging.getLogger(__name__)

interface = "readers"
family = "standard"
name = "viirs_sdr_hdf5"
source_names = ["viirs"]

with import_optional_dependencies(loglevel="info"):
    """Attempt to import a package and print to LOG.info if the import fails."""
    import satpy

VARLIST = {
    "DNB": ["DNB"],
    "IMG": ["I04", "I05"],
    "IMG-Vis": ["I01", "I02", "I03"],
    "MOD": ["M07", "M08", "M10", "M11", "M12", "M13", "M14", "M15", "M16"],
    "MOD-Vis": ["M01", "M02", "M03", "M04", "M05", "M06", "M09"],
}


def bowtie_correction(band, lat, lon):
    """Correct input data for the instrument bowtie effect.

    Correction derived from: Remote Sens. 2016, 8, 79.
    """
    # Unfold
    ord_lat = np.sort(lat, axis=0)
    unfold_idx = np.argsort(lat, axis=0)

    rad_fold = np.take_along_axis(band, unfold_idx, axis=0)
    sort_lon = np.take_along_axis(lon, unfold_idx, axis=0)

    if np.all(np.isnan(rad_fold)):
        LOG.debug("All nan band, no bowtie correction")
        return rad_fold, ord_lat, sort_lon
    if np.all(np.isnan(rad_fold)):
        LOG.debug("All NaNs, no bowtie correction")
        return rad_fold, ord_lat, sort_lon

    res_band = rad_fold.copy()
    # mask and indicies of bad data
    rad_nan_flag = np.isnan(rad_fold)
    nan_mask = np.where(~rad_nan_flag)

    interp = NearestNDInterpolator(np.transpose(nan_mask), rad_fold[nan_mask])
    interp_nans = interp(*np.indices(rad_fold.shape)[:, rad_nan_flag], **{"workers": 4})
    res_band[rad_nan_flag] = interp_nans

    return res_band, ord_lat.astype(np.float64), sort_lon.astype(np.float64)


def call(
    fnames,
    metadata_only=False,
    chans=None,
    area_def=None,
    self_register=False,
    resample=False,
    bowtie=True,
):
    """Read VIIRS SDR hdf5 data products.

    Parameters
    ----------
    fnames : list
        * List of strings, full paths to files
    metadata_only : bool, default=False
        * Return before actually reading data if True
    chans : list of str, default=None
        * List of desired channels (skip unneeded variables as needed).
        * Include all channels if None.
    area_def : pyresample.AreaDefinition, default=None
        * NOT YET IMPLEMENTED
        * Specify region to read
        * Read all data if None.
    self_register : str or bool, default=False
        * NOT YET IMPLEMENTED
        * register all data to the specified dataset id (as specified in the
          return dictionary keys).
        * Read multiple resolutions of data if False.

    Returns
    -------
    dict of xarray.Datasets
        * dictionary of xarray.Dataset objects with required Variables and
          Attributes.
        * Dictionary keys can be any descriptive dataset ids.
        * Conforms to geoips xarray standards, see more in geoips documentation.
    """
    # print("Reading")
    tmp_scn = satpy.Scene(reader="viirs_sdr", filenames=fnames)
    scn_start, scn_end = tmp_scn.start_time, tmp_scn.end_time
    base_fnames = list(map(os.path.basename, fnames))

    full_xr = {}
    if metadata_only:
        # average resolution
        # sensor, plat name
        tmp_scn.load([tmp_scn.available_dataset_names()[0]])
        tmp_attrs = tmp_scn[tmp_scn.available_dataset_names()[0]].attrs
        tmp_xr = xr.Dataset(
            attrs={
                "source_file_name": base_fnames[0],
                "start_datetime": scn_start,
                "end_datetime": scn_end,
                "source_name": tmp_attrs["sensor"],
                "platform_name": tmp_attrs["platform_name"],
                "data_provider": "NOAA",
                "sample_distance_km": 1,
                "interpolation_radius_of_influence": 1000,  # guess!
            }
        )
        tmp_dict = {"METADATA": tmp_xr}
        return tmp_dict

    # trim VARLIST based on channels requested
    if chans:
        tmp_vl = VARLIST.copy()
        for key, val in tmp_vl.items():
            km = [c[:3] in val for c in chans]
            if not any(km):
                VARLIST.pop(key)
            else:
                matches = [s for s in val for c in chans if s in c]
                VARLIST[key] = matches

    # could opimize more
    tmp_coor = {}
    for var in VARLIST:
        tmp_dask = {}
        dataset_ids = [
            idx
            for idx in tmp_scn.available_dataset_ids()
            if idx["name"] in VARLIST[var]
        ]
        if len(dataset_ids) == 0:
            LOG.warning("No variables found in data for viirs_sdr reader.")
            continue

        for d in dataset_ids:
            tmp_scn.load([d])
            full_key = tmp_scn[d].attrs["name"] + tmp_scn[d].attrs[
                "calibration"
            ].capitalize()[:3].replace("Bri", "BT")

            tmp_ma = tmp_scn[d].to_masked_array().data

            # load band
            tmp_scn.load([d])

            lat = tmp_scn[d].area.lats.to_masked_array().data
            lon = tmp_scn[d].area.lons.to_masked_array().data

            # bowtie correction
            if bowtie:
                band_data, band_lat, band_lon = bowtie_correction(tmp_ma, lat, lon)
            else:
                band_data, band_lat, band_lon = tmp_ma, lat, lon

            if "Ref" in full_key:
                from geoips.data_manipulations.corrections import apply_gamma

                # gamma expects data 0-1 range
                # Gamma factor derived from log rule
                # log(mean)/log(0.5) for mean rng 0 to 1
                band_data = apply_gamma(band_data / 100, 1.65) * 100

            tmp_dask |= {full_key: (("dim_0", "dim_1"), band_data)}

        # coordinates
        tmp_coor["latitude"] = (
            ("dim_0", "dim_1"),
            band_lat,
        )
        tmp_coor["longitude"] = (
            ("dim_0", "dim_1"),
            band_lon,
        )

        # sample time to the proper shape (N*48), while lat/lon are ()
        time_range = date_range(
            start=scn_start,
            end=scn_end,
            periods=tmp_coor["latitude"][1].shape[0],
        ).values
        interp_time = np.tile(time_range, (tmp_coor["latitude"][1].shape[1], 1)).T
        tmp_coor["time"] = (("dim_0", "dim_1"), interp_time)

        tmp_attrs = tmp_scn[VARLIST[var][0]].attrs

        cal_params = [
            "satellite_azimuth_angle",
            "satellite_zenith_angle",
            "solar_azimuth_angle",
            "solar_zenith_angle",
        ]

        if var == "DNB":
            cal_params = [
                "dnb_lunar_azimuth_angle",
                "dnb_lunar_zenith_angle",
                "dnb_satellite_azimuth_angle",
                "dnb_satellite_zenith_angle",
                "dnb_solar_azimuth_angle",
                "dnb_solar_zenith_angle",
            ]

        tmp_scn.load(cal_params)
        tmp_cal_params = {
            i.removeprefix("dnb_"): (
                ("dim_0", "dim_1"),
                tmp_scn[i].to_masked_array(),
            )
            for i in cal_params
        }

        if var == "DNB":
            try:
                from lunarref.lib.liblunarref import lunarref

                # this results in the wrong value..
                # np.arccos((tmp_scn["dnb_moon_illumination_fraction"].data/50)-1)

                dnb_geofile = [i for i in fnames if "GDNBO" in os.path.basename(i)][0]
                h5_dnb = h5py.File(dnb_geofile)
                phase_ang = h5_dnb["All_Data/VIIRS-DNB-GEO_All/MoonPhaseAngle"][...]

                lunarref_data = lunarref(
                    tmp_dask["DNBRad"][1],
                    tmp_cal_params["solar_zenith_angle"][1],
                    tmp_cal_params["lunar_zenith_angle"][1],
                    scn_start.strftime("%Y%m%d%H"),
                    scn_start.strftime("%M"),
                    phase_ang,
                )
                lunarref_data = np.ma.masked_less_equal(lunarref_data, -999, copy=False)
                tmp_dask |= {"DNBRef": (("dim_0", "dim_1"), lunarref_data)}
            except ImportError:
                LOG.info("Failed lunarref in viirs reader.  If you need it, build it")

        # problem with sat_za/az values being too high, need to downsample
        # print("Building xarray")
        obs_xr = xr.Dataset(data_vars=tmp_dask)
        coor_xr = xr.Dataset(data_vars=tmp_coor)
        cal_xr = xr.Dataset(data_vars=tmp_cal_params)

        if cal_xr.sizes["dim_0"] > coor_xr.sizes["dim_0"]:
            cal_xr = cal_xr.coarsen(dim_0=2, dim_1=2).mean()

        tmp_xr = xr.merge([obs_xr, coor_xr, cal_xr])

        tmp_xr.attrs = {
            "source_file_name": base_fnames,
            "start_datetime": tmp_scn.start_time,
            "end_datetime": tmp_scn.end_time,
            "source_name": tmp_attrs["sensor"],
            "platform_name": tmp_attrs["platform_name"],
            "data_provider": "NOAA",
            "sample_distance_km": tmp_attrs["resolution"] / 1e3,
            "interpolation_radius_of_influence": 1000,
        }

        full_xr |= {var: tmp_xr}
    full_xr["METADATA"] = xr.Dataset(attrs=tmp_xr.attrs)

    # Geolocation resampling
    if resample and area_def:
        adname = area_def.area_id
        new_shape = area_def.shape

        LOG.info("")
        LOG.info("Getting geolocation information for {}.".format(adname))

        for dtype in full_xr.keys():
            if "latitude" not in full_xr[dtype].variables:
                LOG.info(
                    "No data read for dataset %s, removing from xarray list", dtype
                )
                continue
            fldk_lats = full_xr[dtype]["latitude"]
            fldk_lons = full_xr[dtype]["longitude"]
            # Get just the metadata we need
            geo_metadata = full_xr[dtype].attrs.copy()
            geo_metadata["num_lines"], geo_metadata["num_samples"] = fldk_lats.shape
            geo_metadata["scene"] = "stitched_granules"
            geo_metadata["fnames"] = geo_metadata.pop("source_file_name")
            geo_metadata["roi_factor"] = 5
            lines, samples = get_indexes(geo_metadata, fldk_lats, fldk_lons, area_def)

            index_mask = lines != -999

            new_dim0 = "dim_{:d}".format(new_shape[0])
            new_dim1 = "dim_{:d}".format(new_shape[1])

            for varname in full_xr[dtype].variables.keys():
                new_var = np.full(new_shape, -999.1)
                new_var[index_mask] = full_xr[dtype][varname].values[
                    lines[index_mask], samples[index_mask]
                ]
                # out_var = new_var
                if varname not in cal_params + ["latitude", "longitude"]:
                    # Set values <= -999.9 to NaN so they also get interpolated.
                    new_var[np.where(new_var <= -999.9)] = np.nan
                    # Interpolate missing data.
                    out_var = np.array(DataFrame(new_var).interpolate())
                else:
                    out_var = new_var
                full_xr[dtype][varname] = xr.DataArray(
                    out_var, dims=[new_dim0, new_dim1]
                )
            if "time" in full_xr[dtype].variables:
                time_data = full_xr[dtype]["time"].data
                time_data = np.asarray(to_datetime(time_data.ravel())).reshape(
                    time_data.shape
                )
                full_xr[dtype]["time"].data = time_data

            ll_mask = full_xr[dtype]["latitude"].data >= -90
            for varname in full_xr[dtype].variables.keys():
                full_xr[dtype][varname] = full_xr[dtype][varname].where(ll_mask)

        LOG.interactive("Done with geolocation for {}".format(adname))
        LOG.info("")

    return full_xr



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""GeoIPS CLI "validate" command.

Validates the appropriate plugin based on the arguments provided using the associated
interface's validation mechaninism (interface.plugin_is_valid(plugin_name)).
"""

from importlib.util import spec_from_file_location, module_from_spec
from os.path import exists
from pathlib import Path

import yaml

from geoips.commandline.geoips_command import GeoipsExecutableCommand
from geoips import interfaces


class GeoipsValidate(GeoipsExecutableCommand):
    """Validate Command for validating package plugins."""

    name = "validate"
    command_classes = []

    def add_arguments(self):
        """Add arguments to the validate-subparser fot the Validate Command."""
        self.parser.add_argument(
            "file_path",
            type=str,
            help="File path which represents a GeoIPS Plugin that we want to validate.",
        )
        self.parser.add_argument(
            "plugin_name",
            type=str,
            default=None,
            nargs="?",
            help=(
                "The name of the plugin in the file if applicable. Only useful if your "
                "file is a multi-document yaml file."
            ),
        )

    def __call__(self, args):
        """Validate the appropriate Plugin given the provided arguments.

        Validate the appropriate Plugin based on the arguments provided. This
        acts similar to <geoips_interface>.plugin_is_valid(), but uses the file_path and
        associated interface from the plugin to validate at runtime.
        """
        fpath = Path(args.file_path)
        plugin_name = args.plugin_name
        if not exists(fpath):
            self.parser.error(
                f"Provided filepath '{fpath}' doesn't exist. Provide a valid path.",
            )
        interface, plugin, plugin_name = self.get_interface_and_plugin(
            fpath, plugin_name
        )
        if interface.name == "products":
            is_valid = self.validate_sub_products(interface, fpath, plugin)
        elif interface.interface_type == "yaml_based":
            is_valid = interface.plugin_is_valid(plugin_name)
        else:
            is_valid = interface.plugin_is_valid(
                interface._plugin_module_to_obj(plugin_name, plugin)
            )
        if not is_valid:
            # if it's not valid, report that to the user
            self.parser.error(f"Plugin '{plugin_name}' found at {fpath} is invalid.")
        else:
            # otherwise let them know they're good to go
            print(f"Plugin '{plugin_name}' found at {fpath} is valid.")

    def get_interface_and_plugin(self, fpath, plugin_name=None):
        """Retrieve the interface and plugin associated with the file path given.

        Parameters
        ----------
        fpath: str
            - The file path of the plugin requested to be validated.
        plugin_name: str, default=None
            - If provided and the filepath is a .yaml file, assume this is a
              multi-document yaml file (such as a workflow plugin), and attempt to
              find the plugin matching 'plugin_name' in that document.

        Returns
        -------
        interface: GeoIPS Interface Class
            - The interface associated with the provided plugin.
        plugin: Python Module or Yaml Dictionary
            - The plugin "definition" associated with the file path provided.
        plugin_name: str
            - The name of the plugin
        """
        if fpath.suffix == ".py":
            # module-based plugin
            interface_type = "module_based"
            plugin = self._load_module_from_file(fpath)
        elif fpath.suffix == ".yaml":
            # yaml-based plugin
            interface_type = "yaml_based"
        else:
            self.parser.error(
                "Only '.py' and '.yaml' files are accepted at this time. Try again."
            )
        try:
            # if the module / yaml plugin is missing either interface or name, it's
            # invalid and we need to report the error appropriately
            if interface_type == "module_based":
                interface_name = plugin.interface
                plugin_name = plugin.name
            else:
                # If plugin_name already exists, assume this is a multi-document yaml
                # file and attempt to find the correct plugin within that file
                if plugin_name:
                    docs = yaml.safe_load_all(open(fpath, "r"))
                    plugin = None
                    for doc in docs:
                        if doc["name"] == plugin_name:
                            plugin = doc
                            break
                    # No matching plugin could be found. Raise an error
                    if not plugin:
                        self.parser.error(
                            f"Error: No plugin under name '{plugin_name}' could be "
                            f"in the multi-document yaml plugin at {fpath}."
                        )
                else:
                    with open(fpath, "r") as fo:
                        plugin = yaml.safe_load(fo)
                interface_name = plugin["interface"]
                plugin_name = plugin["name"]
        except AttributeError or KeyError:
            # Report such error.
            err_str = f"Plugin found at {fpath} doesn't have 'interface' and/or "
            err_str += "'name' attribute[s]. This plugin is invalid."
            self.parser.error(err_str)
        # get the correct geoips interface associated with the plugin
        interface = getattr(interfaces, interface_name)
        return interface, plugin, plugin_name

    def _load_module_from_file(self, file_path, module_name=None):
        """Load in a given python module provided a file_path and an optional name."""
        if module_name is None:
            # Generate a unique module name if not provided
            module_name = "module_from_"
            module_name += (
                str(file_path).replace("/", "_").replace(".", "_").replace("\\", "_")
            )

        spec = spec_from_file_location(module_name, file_path)
        module = module_from_spec(spec)
        spec.loader.exec_module(module)
        return module

    def validate_sub_products(self, interface, fpath, plugin):
        """Validate each sub-product plugin found within a products yaml definition.

        If the corresponding interface was found to be a "products" interface, validate
        each sub-product plugin found within the provided yaml products plugin.

        Parameters
        ----------
        interface: GeoIPS Interface Class
            - GeoIPS Products Interface used for validation
        fpath: str
            - The file path of the products plugin to be validated
        plugin: dict
            - Dictionary representing the Products' yaml file provided.

        Returns
        -------
        bool:
            - True or False, where True means that every sub-plugin is valid and False
              means that at least a single sub-plugin was invalid.
        """
        try:
            product_list = plugin["spec"]["products"]
        except KeyError:
            err_str = f"Plugin '{plugin['name']} found at {fpath} is invalid. "
            err_str += "Missing either 'spec' or 'spec['products']' key."
            print(err_str)
            return False

        for subplg in product_list:
            if not interface.plugin_is_valid(subplg["source_names"][0], subplg["name"]):
                return False
        return True



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Read derived surface winds from CYGNSS netcdf data."""

# Python Standard Libraries
import glob
import logging

# Third-Party Libraries
import xarray

# GeoIPS imports
from geoips.xarray_utils.time import get_min_from_xarray_time, get_max_from_xarray_time

LOG = logging.getLogger(__name__)

MS_TO_KTS = 1.94384
DEG_TO_KM = 111.321

interface = "readers"
family = "standard"
name = "cygnss_netcdf"


def call(fnames, metadata_only=False, chans=None, area_def=None, self_register=False):
    """Read CYGNSS derived winds from netcdf data.

    Parameters
    ----------
    fnames : list
        * List of strings, full paths to files
    metadata_only : bool, default=False
        * Return before actually reading data if True
    chans : list of str, default=None
        * List of desired channels (skip unneeded variables as needed).
        * Include all channels if None.
    area_def : pyresample.AreaDefinition, default=None
        * NOT YET IMPLEMENTED
        * Specify region to read
        * Read all data if None.
    self_register : str or bool, default=False
        * NOT YET IMPLEMENTED
        * register all data to the specified dataset id (as specified in the
          return dictionary keys).
        * Read multiple resolutions of data if False.

    Returns
    -------
    dict of xarray.Datasets
        * dictionary of xarray.Dataset objects with required Variables and
          Attributes.
        * Dictionary keys can be any descriptive dataset ids.

    See Also
    --------
    :ref:`xarray_standards`
        Additional information regarding required attributes and variables
        for GeoIPS-formatted xarray Datasets.
    """
    out_dict = {}

    # Single file is full day of data, so only need one.
    fname = fnames[0]

    wind_xarray = xarray.open_dataset(str(fname))
    wind_xarray = wind_xarray.rename(
        {"lat": "latitude", "lon": "longitude", "sample_time": "time"}
    )
    wind_xarray.attrs["source_name"] = "cygnss"
    wind_xarray.attrs["platform_name"] = "cygnss"
    wind_xarray.attrs["data_provider"] = "noaa-nesdis-star-socd-oswt"
    wind_xarray.attrs["source_file_names"] = fnames
    wind_xarray.attrs["interpolation_radius_of_influence"] = 20000
    wind_xarray.attrs["sample_distance_km"] = DEG_TO_KM / 4

    wind_xarray.attrs["start_datetime"] = get_min_from_xarray_time(wind_xarray, "time")
    wind_xarray.attrs["end_datetime"] = get_max_from_xarray_time(wind_xarray, "time")

    out_dict["METADATA"] = wind_xarray[[]]
    if metadata_only:
        return out_dict

    if chans:
        if "wind_speed_kts" in chans:
            wind_xarray["wind_speed_kts"] = wind_xarray["wind_speed"] * MS_TO_KTS
            wind_xarray["wind_speed_kts"].attrs["units"] = "kts"

    ds_var_list = list(wind_xarray.variables.keys())

    if chans:
        for chan in chans:
            if chan not in ds_var_list:
                raise ValueError(f"Variable {chan} not found in dataset.")

        wind_xarray = wind_xarray[chans]

    out_dict["CYGNSS"] = wind_xarray

    return out_dict


# Unit test functions
def get_test_files(test_data_dir):
    """
    Generate a testing xarray dataset from NetCDF files in the test data directory.

    Parameters
    ----------
    test_data_dir : str
        Path to the directory containing test data.

    Returns
    -------
    xarray.Dataset
        The dataset created from the NetCDF test files.

    Raises
    ------
    FileNotFoundError
        If no NetCDF files are found in the specified directory.
    """
    filepath = test_data_dir + "/test_data_cygnss/data/*.nc"
    filelist = glob.glob(filepath)
    if not filelist:
        raise FileNotFoundError("No files found")
    tmp_xr = call(filelist)
    return tmp_xr


def get_test_parameters():
    """Generate test data key for unit testing."""
    return [
        {"data_key": "CYGNSS", "data_var": "wind_speed", "mean": 7.0251474},
    ]



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Interpolation routines from the scipy package."""

# Installed Libraries
import logging

# from matplotlib import cm, colors
import scipy
import numpy

# imports for alphashape-based masking
import alphashape
from skimage.measure import points_in_poly

LOG = logging.getLogger(__name__)

# interface = None indicates to the GeoIPS interfaces that this is not a valid
# plugin, and this module will not be added to the GeoIPS plugin registry.
# This allows including python modules within the geoips/plugins directory
# that provide helper or utility functions to the geoips plugins, but are
# not full GeoIPS plugins on their own.
interface = None


def interp_gaussian_kde(data_lons, data_lats, target_lons, target_lats, vw_method=None):
    """
    Interpolate a given array of non-uniform data using scipy.stats.gaussian_kde.

    This is not finalized.

    Parameters
    ----------
    data_array : numpy.ma.core.MaskedArray
        numpy array of data to interpolate
    data_lons : numpy.ma.core.MaskedArray
        numpy array of longitudes corresponding to original data,
        same shape as data_array
    data_lats : numpy.ma.core.MaskedArray
        numpy array of latitudes corresponding to original data,
        same shape as data_array
    target_lons : numpy.ma.core.MaskedArray
        2d numpy array of desired longitudes
    target_lats : numpy.ma.core.MaskedArray
        2d numpy array of desired latitudes
    bw_method : str
        Bandwidth selection method (see scipy.stats.gaussian_kde)|

    See Also
    --------
    scipy.stats.gaussian_kde
    """
    from scipy import stats

    interp_data = None

    positions = numpy.vstack([target_lons.ravel(), target_lats.ravel()])
    values = numpy.vstack([data_lons, data_lats])
    kernel = stats.gaussian_kde(values)
    kernel(positions)

    return interp_data


def interp_griddata(
    data_array,
    data_lons,
    data_lats,
    min_gridlon,
    max_gridlon,
    min_gridlat,
    max_gridlat,
    numx_grid,
    numy_grid,
    method="linear",
):
    """
    Interpolate a given array of non-uniform data to a specified grid.

    Uses scipy.interpolate.griddata

    Parameters
    ----------
    data_array : numpy.ma.core.MaskedArray
        numpy array of original data to be interpolated
    data_lons : numpy.ma.core.MaskedArray
        numpy array of longitudes corresponding to original data,
        same shape as data_array
    data_lats : numpy.ma.core.MaskedArray
        numpy array of latitudes corresponding to original data,
        same shape as data_array
    min_gridlon : float
        minimum desired lon for the output grid
          * -180.0 < min_gridlon < 180.0
    max_gridlon : float
        maximum desired lon for the output grid
          * -180.0 < max_gridlon < 180.0
    min_gridlat : float
        minimum desired lat for the output grid
          * -90.0 < min_gridlat < 90.0
    max_gridlat : float
        maximum desired lat for the output grid
          * -90.0 < max_gridlat < 90.0
    numx_grid : int
        number desired longitude points in the output grid
    numy_grid : int
        number desired latitude points in the output grid

    Other Parameters
    ----------------
    method: str, default='linear'
        A string specifying the interpolation method to use
        for scipy.interpolate.griddata.
        One of 'nearest', 'linear' or 'cubic'
    """
    # make sure that 'longitude' is in 0-360 deg if the sector is crosssing the dateline

    if min_gridlon > 0 and max_gridlon < 0:
        data_lons = numpy.where(data_lons > 0, data_lons, 360 + data_lons)
        max_gridlon = 360 + max_gridlon

    # data_lons = numpy.where(data_lons >0,data_lons,360+data_lons)

    if len(data_lons.shape) > 1:
        data_lons = data_lons.flatten()
        data_lats = data_lats.flatten()
        data_array = data_array.flatten()

    if hasattr(data_array, "mask"):
        inds = numpy.ma.where(data_array)
        data_lons = data_lons[inds]
        data_lats = data_lats[inds]
        data_array = data_array[inds]
    # ocnvert negative longituge into 0-360 if needed
    # if min_gridlon <0:
    #    min_gridlon = 360 + min_gridlon
    # if max_gridlon <0:
    #    max_gridlon = 360 + max_gridlon

    xx = numpy.linspace(min_gridlon, max_gridlon, int(numx_grid))
    yy = numpy.linspace(max_gridlat, min_gridlat, int(numy_grid))
    gridlons, gridlats = numpy.meshgrid(xx, yy)

    # Free up memory ??
    xx = 0
    yy = 0

    interp_data = scipy.interpolate.griddata(
        (data_lats, data_lons), data_array, (gridlats, gridlons), method=method
    )
    list_of_lons = numpy.reshape(gridlons, -1)
    list_of_lats = numpy.reshape(gridlats, -1)

    # Free up memory ??
    gridlons = 1
    gridlats = 1

    # Zip data_lons and data_lats and compute alphashape
    datapts_2d = list(zip(data_lons.data, data_lats.data))
    returned_polygons = alphashape.alphashape(datapts_2d, 1.0)
    # If multiple polygons are returned, choose the one with the largest area
    if returned_polygons.geom_type == "MultiPolygon":
        alpha_shape = max(returned_polygons.geoms, key=lambda a: a.area)
    elif returned_polygons.geom_type == "Polygon":
        alpha_shape = returned_polygons

    interp_data = numpy.ma.masked_invalid(interp_data)
    interp_data = numpy.ma.masked_less(interp_data, data_array.min())
    interp_data = numpy.ma.masked_greater(interp_data, data_array.max())
    input_shape_interp_data = interp_data.shape

    zipped_lons_lats = numpy.asarray(list(zip(list_of_lons, list_of_lats)))
    mask_exterior_coords_xy = numpy.asarray(list(zip(*alpha_shape.exterior.coords.xy)))
    inside_hull_mask = numpy.reshape(
        points_in_poly(zipped_lons_lats, mask_exterior_coords_xy),
        input_shape_interp_data,
    )

    interp_data = numpy.ma.masked_where(
        numpy.logical_not(inside_hull_mask), interp_data
    )

    return interp_data



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Unit test for GeoIPS CLI `test linting` command.

See geoips/commandline/ancillary_info/cmd_instructions.yaml for more information.
"""

import pytest

from tests.unit_tests.commandline.cli_top_level_tester import BaseCliTest


class TestGeoipsTestLinting(BaseCliTest):
    """Unit Testing Class for Test Linting Command."""

    @property
    def command_combinations(self):
        """A list of every possible call signature for the GeoipsTestLinting command.

        This includes failing cases as well.
        """
        if not hasattr(self, "_cmd_list"):
            base_args = ["geoips", "test", "linting"]
            self._cmd_list = [base_args]
            # select a small random amount of tests to call via geoips run
            for pkg_name in self.plugin_package_names:
                self._cmd_list.append(base_args + ["-p", pkg_name])
            # Add argument list to retrieve help message
            self._cmd_list.append(base_args + ["-h"])
            # Add argument list with non existent package
            self._cmd_list.append(base_args + ["-p", "non_existent_package"])
        return self._cmd_list

    def check_error(self, args, error):
        """Ensure that the 'geoips test linting ...' error output is correct.

        Parameters
        ----------
        args: 2D list of str
            - The arguments used to call the CLI (expected to fail)
        error: str
            - Multiline str representing the error output of the CLI call
        """
        editable = self.assert_non_editable_error_or_wrong_package(args, error)
        if editable:
            # An error occurred using args. Assert that args is not valid and check the
            # output of the error.
            assert "To use, type `geoips test linting -p <package_name>`" in error

    def check_output(self, args, output):
        """Ensure that the 'geoips test linting ...' successful output is correct.

        Parameters
        ----------
        args: 2D list of str
            - The arguments used to call the CLI
        output: str
            - Multiline str representing the output of the CLI call
        """
        # The args provided are valid, so test that the output is actually correct
        if "-h" in args:
            assert "To use, type `geoips test linting -p <package_name>`" in output
        else:
            # Checking that output from geoips test linting command reports succeeds
            for linter in ["bandit", "black", "flake8"]:
                assert f"CALLING TEST:\n{linter}" in output
                assert f"TEST COMPLETE {linter}" in output


test_sub_cmd = TestGeoipsTestLinting()


@pytest.mark.parametrize(
    "args",
    test_sub_cmd.command_combinations,
    ids=test_sub_cmd.generate_id,
)
def test_command_combinations(monkeypatch, args):
    """Test all 'geoips test linting ...' commands.

    This test covers every valid combination of commands for the 'geoips test linting'
    command. We also test invalid commands, to ensure that the proper help documentation
    is provided for those using the command incorrectly.

    Parameters
    ----------
    args: 2D array of str
        - List of arguments to call the CLI with (ie. ['geoips', 'test', 'linting'])
    """
    test_sub_cmd.test_command_combinations(monkeypatch, args)



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Simple test script to run "test_interface" for each interface."""
from pprint import pformat
import traceback
import json

from geoips import interfaces
from geoips.interfaces.base import BaseInterface
from geoips.commandline.log_setup import setup_logging
from geoips.commandline.args import get_argparser, check_command_line_args
from geoips.utils.context_managers import import_optional_dependencies

FAILED_INTERFACE_HEADER_PRE = "FAILED: issues found within "
FAILED_INTERFACE_HEADER_POST = "interface:"


def main():
    """Script to test all plugins in all interfaces."""
    supported_args = ["logging_level"]

    # Use standard GeoIPS argparser - use standard formatting for standard args.
    parser = get_argparser(
        arglist=supported_args,
        description="Test all interfaces within the current geoips instantiation.",
    )
    # Add our specific "interfaces" argument.
    parser.add_argument(
        "interfaces",
        nargs="*",
        default=None,
        help="""List of interfaces to include.  If None, include all interfaces.""",
    )

    # Get the dictionary of command line args.
    COMMAND_LINE_ARGS = parser.parse_args().__dict__
    # Check included arguments for appropriate formatting / type.
    check_command_line_args(supported_args, COMMAND_LINE_ARGS)

    # Setup logging at the requested logging level.
    LOG = setup_logging(logging_level=COMMAND_LINE_ARGS["logging_level"])

    failed_interfaces = []
    failed_plugins = []
    failed_registries = []

    successful_interfaces = []
    successful_plugins = []
    successful_registries = []

    failed_errors = []
    failed_tracebacks = []

    # If "interfaces" is not specified command line, then include plugins for
    # all interfaces.
    if not COMMAND_LINE_ARGS["interfaces"]:
        curr_interfaces = interfaces.__dict__.values()
    else:
        curr_interfaces = [
            interfaces.__dict__[curr_interface]
            for curr_interface in COMMAND_LINE_ARGS["interfaces"]
        ]

    out_dicts = {}

    # Plugin registry was previously being tested during runtime.
    # We want to avoid catastrophic failure at runtime for a single
    # bad plugin, so ensure these are validated in testing, and allow
    # bad plugins to get through at runtime.

    # NOTE: The section below will only be used if the user has pytest installed. We
    # still should figure out where and when we'd like to validate the plugin
    # registries. Currently validation of the registires invokes pytest, which is not a
    # required dependency of GeoIPS. Calls 'plg_reg.validate_all_registries()' and
    # 'plg_reg.validate_registry(pkg_plugins, reg_path)' are no longer on the plugin
    # registry class, rather, they can be found under
    # 'geoips/tests/unit_tests/plugin_registries/test_plugin_registries.py' -->
    # PluginRegistryValidator

    with import_optional_dependencies(loglevel="interactive"):
        """Attempt to import the pytest-based PluginRegistryValidator.

        If this import works, validate the registries for testing purposes. See
        'tests.unit_tests.plugin_registries.test_plugin_registries' for more
        information.
        """
        from tests.unit_tests.plugin_registries.test_plugin_registries import (
            PluginRegistryValidator,
        )
        from geoips.errors import PluginRegistryError

        plg_reg = PluginRegistryValidator()
        try:
            LOG.interactive("Testing all registries...")
            plg_reg.validate_all_registries()
            successful_registries += ["all"]
        except PluginRegistryError as resp:
            failed_registries += ["all"]
            failed_tracebacks += [
                f"\n\n\n{FAILED_INTERFACE_HEADER_PRE}\n\n{traceback.format_exc()}"
            ]
            failed_errors += [f"\n\n\n{FAILED_INTERFACE_HEADER_PRE}\n\n {str(resp)}"]

        # Now test each registry file individually.
        for reg_path in plg_reg.registry_files:
            pkg_plugins = json.load(open(reg_path, "r"))
            LOG.interactive(f"Testing registry: {reg_path}...")
            try:
                plg_reg.validate_registry(pkg_plugins, reg_path)
                successful_registries += [reg_path]
            except PluginRegistryError as resp:
                failed_registries += [reg_path]
                failed_tracebacks += [
                    f"\n\n\n{FAILED_INTERFACE_HEADER_PRE} '{reg_path}'\n\n"
                    f"{traceback.format_exc()}"
                ]
                failed_errors += [
                    f"\n\n\n{FAILED_INTERFACE_HEADER_PRE} '{reg_path}'\n\n{str(resp)}"
                ]

    # Loop through all requested interfaces, fully testing all plugins in each.
    # Collect output in lists, so we can fully print everything at the end before
    # raising an exception on error.
    for curr_interface in curr_interfaces:
        # Do not test "BaseInterface"
        if (type(curr_interface) is BaseInterface) or not isinstance(
            curr_interface, BaseInterface
        ):
            continue

        LOG.info("")
        LOG.interactive(f"Testing {curr_interface.name}...")

        # Open all the interfaces (not just checking call signatures)
        # This returns a dictionary of all sorts of stuff.
        try:
            out_dict = curr_interface.test_interface()
            out_dicts[curr_interface.name] = out_dict
        except Exception as resp:
            LOG.info(traceback.format_exc())
            failed_plugins += [curr_interface.name]
            # Collect tracebacks and errors in lists, will print everything at the end
            # prior to raising the exception.
            failed_tracebacks += [
                f"\n\n\n{FAILED_INTERFACE_HEADER_PRE} '{curr_interface.name}' "
                f"{FAILED_INTERFACE_HEADER_POST}\n\n"
                f"{traceback.format_exc()}"
            ]
            failed_errors += [
                f"\n\n\n{FAILED_INTERFACE_HEADER_PRE} '{curr_interface.name}' "
                f"{FAILED_INTERFACE_HEADER_POST}\n\n"
                f"{str(resp)}"
            ]

    # Collect lists of the successful and failed plugins (without actual errors)
    for intname in out_dicts:
        LOG.info("%s", pformat(out_dict))
        out_dict = out_dicts[intname]
        if out_dict["all_valid"] is True:
            successful_interfaces += [intname]
        else:
            failed_interfaces += [intname]
        for modname in out_dict["validity_check"]:
            if not out_dict["validity_check"][modname]:
                failed_plugins += [f"{intname} on {modname}"]
            else:
                successful_plugins += [f"{intname} on {modname}"]

    # Print all failed plugin tracebacks to interactive log level.
    for failed_traceback in failed_tracebacks:
        LOG.interactive(failed_traceback)
        LOG.interactive("")

    # Only include full list of successful plugins at LOG.info level, not interactive.
    for curr_plugin in successful_plugins:
        LOG.info(f"SUCCESSFUL PLUGIN {curr_plugin}")

    # Print successful registries to interactive log level.
    for curr_successful in successful_registries:
        LOG.interactive(f"SUCCESSFUL REGISTRY {curr_successful}")

    # Print successful interfaces to interactive log level.
    for curr_successful in successful_interfaces:
        LOG.interactive(f"SUCCESSFUL INTERFACE {curr_successful}")

    # Print failed plugins to interactive log level.
    for curr_failed in failed_plugins:
        LOG.interactive(f"FAILED PLUGIN IN {curr_failed}")

    # Print failed registries to interactive log level.
    for curr_failed in failed_registries:
        LOG.interactive(f"FAILED REGISTRY IN {curr_failed}")

    # Print failed interfaces to interactive log level.
    for curr_failed in failed_interfaces:
        LOG.interactive(f"FAILED INTERFACE VALIDITY CHECK DICT {curr_failed}")

    # Print actual failed plugin errors to interactive log level.
    for failed_error in failed_errors:
        LOG.interactive(failed_error)
        LOG.interactive("")

    # Now, after reporting all errors individually, raise an exception if there were
    # any failed interfaces/plugins.  Include list of interfaces/plugins that had
    # failures within the exception string.
    if (
        len(failed_interfaces) > 0
        or len(failed_plugins) > 0
        or len(failed_registries) > 0
    ):
        failed_headers = ""
        for failed_plugin in failed_plugins:
            failed_headers += (
                f"\n  {FAILED_INTERFACE_HEADER_PRE} '{failed_plugin}' "
                f"{FAILED_INTERFACE_HEADER_POST}"
            )
        for failed_registry in failed_registries:
            failed_headers += f"\n  {FAILED_INTERFACE_HEADER_PRE} '{failed_registry}' "
        raise TypeError(
            f"Failed validity checks on registries:\n"
            f"{pformat(failed_registries)}.\n"
            f"Failed validity checks on plugins:\n"
            f"{pformat(failed_plugins)}.\n"
            "\nAdditional information on these errors can be found above, "
            "under headers:\n"
            f"{failed_headers}"
        )


if __name__ == "__main__":
    main()



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""GeoIPS CLI "describe" command.

Retrieves the appropriate family/interface/package/plugin based on the args provided.
"""

from importlib import metadata, resources, import_module

import yaml

from geoips.commandline.geoips_command import (
    CommandClassFactory,
    GeoipsCommand,
    GeoipsExecutableCommand,
)
from geoips.create_plugin_registries import format_docstring
from geoips import interfaces


class GeoipsDescribeArtifact(GeoipsExecutableCommand):
    """Command which returns information describing a GeoIPS artifact.

    Where this artifact is one of ['interface', 'plugin', 'family'.]

    This is called via `geoips describe <interface_name> <opt_args>`. Data included when
    calling this command is shown below, outputted in a yaml-based format.

    * Interface:
        * Command Signature:
            * `geoips describe <interface_name>`
        * Artifact Listing:
            * `geoips list interfaces`
        * output_info:
            * Absolute Path
            * Docstring
            * Interface Name
            * Interface Type
            * Supported Families

    * Plugin:
        * Command Signature:
            * `geoips describe <interface_name> <plugin_name>`
        * Artifact Listing:
            * `geoips list plugins <-p> <package_name>`
        * Output Info:
            * Docstring
            * Family Name
            * GeoIPS Package
            * Interface Name
            * Plugin Type
            * call_sig / source_names / Product Defaults (dependent on Plugin Type)
            * Relative Path

    * Family:
        * Command Signature:
            * `geoips describe <interface_name> family <family_name>`
        * Artifact Listing:
            * `geoips list interfaces --columns interface supported_families`
        * Output Info:
            * Docstring
            * Family Name
            * Family Path
            * Interface Name
            * Interface Type
            * Required Arguments / Schema
    """

    name = "interface"
    command_classes = []

    def add_arguments(self):
        """Add arguments to the describe-subparser for the describe Interface cmd."""
        self.parser.add_argument(
            "plugin_name",
            type=str,
            default=None,
            nargs="?",
            help="GeoIPS Plugin to select from the provided interface.",
        )
        self.parser.add_argument(
            "family_name",
            type=str,
            default=None,
            nargs="?",
            choices=getattr(interfaces, self.name.replace("-", "_")).supported_families,
            help="GeoIPS Family to select from the provided interface.",
        )
        pass

    def __call__(self, args):
        """CLI 'geoips describe <interface_name>' command.

        This occurs when a user has requested a interface in the manner shown above.
        Outputs to the teriminal the following data in a dictionary format if available.

        Printed to Terminal
        -------------------
        yaml-based output: dict
            - Absolute Path
            - Docstring
            - Interface Name
            - Interface Type
            - Supported Families

        Parameters
        ----------
        args: Argparse Namespace()
            - The list argument namespace to parse through
        """
        if (
            args.plugin_name
            and args.plugin_name != "family"
            and args.plugin_name != "fam"
        ):
            self.describe_plugin(args)
        elif (
            args.plugin_name == "family" or args.plugin_name == "fam"
        ) and args.family_name:
            self.describe_family(args)
        elif args.plugin_name is None and args.family_name is None:
            self.describe_interface()
        else:
            self.parser.error(
                "Invalid command ran for 'geoips describe <interface_name>. Please run "
                "'geoips describe -h' for more information on how to run this command."
            )

    def describe_plugin(self, args):
        """CLI 'geoips describe <interface_name> <plugin_name>' command.

        This occurs when a user has requested a plugin in the manner shown above.
        Outputs to the teriminal the following data in a dictionary format if available.

        Printed to Terminal
        -------------------
        yaml-based output: dict
            - Docstring
            - Family Name
            - GeoIPS Package
            - Interface Name
            - Plugin Type
            - call_sig / source_names / Product Defaults (dependent on Plugin Type)
            - Relative Path

        Parameters
        ----------
        args: Argparse Namespace()
            - The list argument namespace to parse through
        """
        interface_name = self.name.replace("-", "_")
        plugin_name = args.plugin_name
        try:
            interface = getattr(interfaces, interface_name)
        except AttributeError:
            self.parser.error(
                f"Interface: {interface_name} doesn't exist. Provide a valid interface."
            )
        # If plugin_name is not None, then the user has requested a plugin within
        # an interface, rather than the interface itself
        interface_registry = interface.plugin_registry.registered_plugins[
            interface.interface_type
        ][interface.name]
        # Ensure the provided plugin exists within the interface's plugin registry
        self.ensure_plugin_exists(interface.name, interface_registry, plugin_name)
        if interface.name == "products":
            source_name, plugin_name = plugin_name.split(".", 1)
            plugin_entry = interface_registry[source_name][plugin_name]
            self._output_dictionary_highlighted(plugin_entry)
        else:
            plugin_entry = interface_registry[plugin_name]
            self._output_dictionary_highlighted(plugin_entry)

    def describe_family(self, args):
        """CLI 'geoips describe <interface_name> family <family_name>' command.

        This occurs when a user has requested a family in the manner shown above.
        Outputs to the teriminal the following data in a dictionary format if available.

        Printed to Terminal
        -------------------
        yaml-based output: dict
            - Docstring
            - Family Name
            - Family Path
            - Interface Name
            - Interface Type
            - Required Arguments / Schema

        Parameters
        ----------
        args: Argparse Namespace()
            - The list argument namespace to parse through
        """
        interface_name = self.name.replace("-", "_")
        family_name = args.family_name
        try:
            interface = getattr(interfaces, interface_name)
        except AttributeError:
            self.parser.error(
                f"Interface: {interface_name} doesn't exist. Provide a valid interface."
            )
        interface_type = interface.interface_type
        supported_families = interface.supported_families
        if family_name not in supported_families:
            # If the family name is not one of the interface's supported family
            # members, raise an error
            err_str = f"Error: Family: `{family_name}` is not within Interface: "
            err_str += f"`{interface_name}` supported families: `{supported_families}`"
            self.parser.error(err_str)
        if interface_type == "module_based":
            docstring = "Not Implemented."
            family_path = str(
                resources.files("geoips")
                / f"interfaces/{interface_type}/{interface_name}.py"
            )
            family_args_or_schema = interface.required_args[family_name]
        else:
            family_path = str(
                resources.files("geoips")
                / f"schema/{interface_name}/{family_name}.yaml"
            )
            with open(family_path, "r") as fo:
                family_args_or_schema = yaml.safe_load(fo)
            if "description" in list(family_args_or_schema.keys()):
                family_args_or_schema["description"] = format_docstring(
                    family_args_or_schema["description"],
                )
                docstring = family_args_or_schema["description"]
            else:
                docstring = "Not Implemented."
        family_entry = {
            "Interface Name": interface_name,
            "Interface Type": interface_type,
            "Family Name": family_name,
            "Required Args / Schema": family_args_or_schema,
            "Docstring": docstring,
            "Family Path": family_path,
        }
        self._output_dictionary_highlighted(family_entry)

    def describe_interface(self):
        """CLI 'geoips describe <interface_name>' command.

        This occurs when a user has requested a interface in the manner shown above.
        Outputs to the teriminal the following data in a dictionary format if available.

        Printed to Terminal
        -------------------
        yaml-based output: dict
            - Absolute Path
            - Docstring
            - Interface Name
            - Interface Type
            - Supported Families
        """
        interface_name = self.name.replace("-", "_")
        try:
            interface = getattr(interfaces, interface_name)
        except AttributeError:
            self.parser.error(
                f"Interface: {interface_name} doesn't exist. Provide a valid interface."
            )

        geoips_pkg_path = resources.files("geoips")
        interface_path = str(
            geoips_pkg_path
            / f"interfaces/{interface.interface_type}/{interface.name}.py"
        )
        interface_entry = {
            "Interface": interface.name,
            "Interface Type": interface.interface_type,
            "Docstring": format_docstring(interface.__doc__),
            "Absolute Path": interface_path,
            "Supported Families": interface.supported_families,
        }
        self._output_dictionary_highlighted(interface_entry)

    def ensure_plugin_exists(self, interface_name, interface_registry, plugin_name):
        """Ensure that the given plugin exists within an interface's plugin registry.

        If the plugin is not found within the interface's registry, raise a KeyError,
        otherwise, just return.

        Parameters
        ----------
        interface_name: str
            - The name of the selected GeoIPS Interface
        interface_registry: dict
            - The plugin registry for the selected GeoIPS Interface
        plugin_name: str
            - The name of the plugin from the selected interface
        """
        if interface_name == "products":
            if "." not in plugin_name:
                err_str = (
                    "Product plugins must be retrieved via `<source_name>."
                    f"<plugin_name>`. Requested {plugin_name} doesn't match that."
                )
                raise KeyError(err_str)
            source_name, plugin_name = plugin_name.split(".", 1)
            if plugin_name not in interface_registry[source_name].keys():
                raise KeyError(
                    f"{plugin_name} not found under Products {source_name} entry."
                )
        elif plugin_name not in interface_registry.keys():
            self.parser.error(
                f"{plugin_name} doesn't exist within Interface {interface_name}."
            )


class GeoipsDescribePackage(GeoipsExecutableCommand):
    """Describe Command which retrieves information about a certain GeoIPS Package.

    This is called via `geoips describe package <interface_name>`. Data included when
    calling this command is shown below, outputted in a yaml-based format.
    """

    name = "package"
    command_classes = []

    def add_arguments(self):
        """Add arguments to the describe-subparser for the describe Package Command."""
        self.parser.add_argument(
            "package_name",
            type=str.lower,
            default="geoips",
            choices=self.plugin_package_names,
            help="GeoIPS Package to retrieve.",
        )

    def __call__(self, args):
        """CLI 'geoips describe package <package_name>' command.

        This occurs when a user has requested a package in the manner shown above.
        Outputs to the teriminal the following data in a dictionary format if available.

        Printed to Terminal
        -------------------
        yaml-based output: dict
            - Docstring
            - GeoIPS Package
            - Package Path
            - Source Code
            - Version Number

        Parameters
        ----------
        args: Argparse Namespace()
            - The list argument namespace to parse through
        """
        package_name = args.package_name
        package_path = str(resources.files(package_name))

        docstring = import_module(package_name).__doc__
        package_entry = {
            "GeoIPS Package": package_name,
            "Docstring": format_docstring(docstring, use_regex=False),
            "Package Path": package_path,
            "Source Code": f"{self.github_org_url}{package_name}",
            "Version Number": metadata.version(package_name),
        }
        self._output_dictionary_highlighted(package_entry)


class GeoipsDescribe(GeoipsCommand):
    """Top-Level Describe Command Class for retrieving info about GeoIPS Artifacts."""

    name = "describe"

    generated_classes = []
    for int_name in sorted(interfaces.__all__):
        generated_classes.append(
            CommandClassFactory(
                GeoipsDescribeArtifact,
                int_name.replace("_", "-"),
            ).generated_class
        )

    command_classes = generated_classes + [GeoipsDescribePackage]
